import uuid
import vdomr as vd
import os
import json

source_path=os.path.dirname(os.path.realpath(__file__))

class ElectrodeGeometryWidget(vd.Component):
    def __init__(self,*,recording):
        vd.Component.__init__(self)
        self._recording=recording
        self._div_id='ElectrodeGeometryWidget-'+str(uuid.uuid4())

        vd.devel.loadBootstrap()
        vd.devel.loadCss(url='https://stackpath.bootstrapcdn.com/font-awesome/4.7.0/css/font-awesome.min.css')
        vd.devel.loadJavascript(path=source_path+'/dist/main.js')
        vd.devel.loadJavascript(path=source_path+'/dist/d3.v5.min.js')
        vd.devel.loadJavascript(path=source_path+'/dist/jquery-3.3.1.min.js')

        js="window.sfdata=window.sfdata||{}; window.sfdata['electrode_geometry_{div_id}']={obj}"
        js=self._div_id.join(js.split('{div_id}'))
        obj=self._get_geom()
        js=json.dumps(obj).join(js.split('{obj}'))
        vd.devel.loadJavascript(js=js)
    def _get_geom(self):
        RX=self._recording
        electrodes=[]
        for ch in RX.getChannelIds():
            location=RX.getChannelProperty(channel_id=ch,property_name='location')
            electrodes.append(
                dict(
                    channel_id=ch,
                    location=list(location)
                )
            )
        obj=dict(electrodes=electrodes)
        return obj
    def render(self):
        div=vd.div(id=self._div_id)
        js="""
        let W=new window.GeomWidget();
        let X=window.sfdata['electrode_geometry_{z_id}']
        console.log(X);
        W.setGeometry(X);
        //W.setSize(300,300)
        $('#{div_id}').empty();
        $('#{div_id}').append(W.div());
        W.update();
        """
        js=self._div_id.join(js.split('{div_id}'))
        vd.devel.loadJavascript(js=js,delay=1)
        return div


import vdomr as vd
from .tabbar import TabBar

class HighlightBox(vd.Component):
  def __init__(self):
    vd.Component.__init__(self)
    self._highlight=True
  def setHighlight(self,val):
    if self._highlight==val:
        return
    self._highlight=val
    self.refresh()
  def render(self):
    style0=dict(width='100%',height='100%',position='absolute')
    style0['z-index']='-1'
    if self._highlight:
        style0['border']='solid 2px black'
    return vd.div(style=style0)

class ViewHolder(vd.Component):
  def __init__(self,child, *, name=''):
    vd.Component.__init__(self)
    self._child=child
    self._name=name
  def name(self):
    return self._name
  def child(self):
    return self._child
  def render(self):
    return vd.div(self._child)

class Container(vd.Component):
  def __init__(self,*args,position,size,position_mode='absolute',style=dict()):
    vd.Component.__init__(self)
    self._children=list(args)
    self._position=position
    self._size=size
    self._position_mode=position_mode
    self._style=style
  def render(self):
    style=self._style
    style['position']=self._position_mode
    style['width']='{}px'.format(self._size[0])
    style['height']='{}px'.format(self._size[1])
    style['left']='{}px'.format(self._position[0])
    style['top']='{}px'.format(self._position[1])
    return vd.div(
        self._children,
        style=style
    )

class ViewContainerContent(vd.Component):
  def __init__(self):
    vd.Component.__init__(self)
    self._child = None
  def setChild(self, child):
    self._child = child
    self.refresh()
  def render(self):
    if not self._child:
      return vd.div('No content')
    return self._child

class ViewContainer(vd.Component):
  def __init__(self):
    vd.Component.__init__(self)
    self._views=dict()
    self._fiew_holders=dict()
    self._content=ViewContainerContent()
    self._last_id_num=0
    self._size=(0,0)
    self._tab_bar=TabBar()
    self._tab_bar.onCurrentTabChanged(self._on_current_tab_changed)
    self._tab_bar.onTabRemoved(self._on_tab_removed)
    self._click_handlers=[]
    self._highlight_box=HighlightBox()
    self._poll()
  def onClick(self,handler):
    self._click_handlers.append(handler)
  def setSize(self,size):
    self._size=size
    for view in self._views.values():
      self._update_view_size(view)
  def addView(self,view, *, name=''):
    id_num = self._last_id_num+1
    self._last_id_num = self._last_id_num + 1
    view_id = 'view_{}'.format(id_num)

    self._update_view_size(view)
    holder=ViewHolder(view, name=name)
    self._views[view_id]=view
    self._fiew_holders[view_id] = holder
    self._tab_bar.addTab(view_id, view.tabLabel())

    # self.refresh()
  def findView(self, *, name):
    for vf in self._fiew_holders.values():
      if vf.name() == name:
        return vf.child()
    return None
  def setCurrentView(self, view):
    for id, v in self._views.items():
      if v == view:
        self._tab_bar.setCurrentTab(id=id)
        return

  def setHighlight(self,val):
    self._highlight_box.setHighlight(val)

  def _update_view_size(self, view):
    view.setSize((self._size[0]-self._tab_bar.height()-10,self._size[1]-self._tab_bar.height()-10))

  def _on_current_tab_changed(self):
    id = self._tab_bar.currentTabId()
    if id:
      holder = self._fiew_holders[id]
      self._content.setChild(holder)
    else:
      self._content.setChild(None)

  def _on_tab_removed(self, id):
    if id in self._views:
      view = self._views[id]
      del self._views[id]
      del self._fiew_holders[id]
      if hasattr(view, 'cleanup'):
        (getattr(view, 'cleanup'))()

  def currentView(self):
    id = self._tab_bar.currentTabId()
    if id is None:
      return None
    if id not in self._views:
      return None
    return self._views[id]
  def _on_click(self):
    for handler in self._click_handlers:
      handler()
  def _poll(self):
    for id, v in self._views.items():
      label = v.tabLabel()
      self._tab_bar.setTabLabel(id, label)
      if hasattr(v, 'updateTitle'):
        v.updateTitle()
    vd.set_timeout(self._poll, 2)
  def render(self):
    style_outer=dict(width='100%',height='100%',position='absolute')
    style_content=dict(left='5px',right='5px',top='{}px'.format(5+self._tab_bar.height()),bottom='5px',position='absolute')
    onclick=self._on_click
    return vd.div(self._highlight_box,vd.div(self._content, id='content-'+self.componentId(),style=style_content),self._tab_bar,style=style_outer,onclick=onclick)



import os
import sys

# def append_to_path(dir0): # A convenience function
#     if dir0 not in sys.path:
#         sys.path.append(dir0)
# append_to_path(os.getcwd()+'/..')
from spikesorters import MountainSort4, SpykingCircus, YASS, KiloSort, IronClust
from spikeforest import example_datasets
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
import tempfile
import shutil
from mountaintools import client as mt
import pytest


def setup_module(module):
    # Use this so we can download containers as needed
    mt.configDownloadFrom(['spikeforest.spikeforest2'])

    # Check if singularity is installed
    retval = os.system('singularity --version')
    assert retval == 0, 'Singularity is not installed'


def teardown_module(module):
    pass


def setup_function(function):
    pass


def teardown_function(function):
    pass


@pytest.mark.slow
@pytest.mark.mountainsort4
def test_mountainsort4(tmpdir):
    tmpdir = str(tmpdir)

    rx, sx = example_datasets.toy_example1()
    SFMdaRecordingExtractor.writeRecording(
        recording=rx, save_path=tmpdir+'/recording')
    SFMdaSortingExtractor.writeSorting(
        sorting=sx, save_path=tmpdir+'/recording/firings_true.mda')

    MountainSort4.execute(
        recording_dir=tmpdir+'/recording',
        firings_out=tmpdir+'/firings.mda',
        detect_sign=-1,
        adjacency_radius=50,
        _container='default'
    )
    assert os.path.exists(tmpdir+'/firings.mda')


@pytest.mark.slow
@pytest.mark.spyking_circus
def test_spyking_circus(tmpdir):
    tmpdir = str(tmpdir)

    rx, sx = example_datasets.toy_example1()
    SFMdaRecordingExtractor.writeRecording(
        recording=rx, save_path=tmpdir+'/recording')
    SFMdaSortingExtractor.writeSorting(
        sorting=sx, save_path=tmpdir+'/recording/firings_true.mda')

    SpykingCircus.execute(
        recording_dir=tmpdir+'/recording',
        firings_out=tmpdir+'/firings.mda',
        detect_sign=-1,
        adjacency_radius=50,
        _container='default'
    )
    assert os.path.exists(tmpdir+'/firings.mda')


@pytest.mark.slow
@pytest.mark.yass
def test_yass(tmpdir):
    tmpdir = str(tmpdir)

    rx, sx = example_datasets.toy_example1()
    SFMdaRecordingExtractor.writeRecording(
        recording=rx, save_path=tmpdir+'/recording')
    SFMdaSortingExtractor.writeSorting(
        sorting=sx, save_path=tmpdir+'/recording/firings_true.mda')

    YASS.execute(
        recording_dir=tmpdir+'/recording',
        firings_out=tmpdir+'/firings.mda',
        detect_sign=-1,
        adjacency_radius=50,
        _container='default'
    )
    assert os.path.exists(tmpdir+'/firings.mda')


@pytest.mark.exclude
@pytest.mark.kilosort
def test_kilosort(tmpdir):
    tmpdir = str(tmpdir)

    rx, sx = example_datasets.toy_example1()
    SFMdaRecordingExtractor.writeRecording(
        recording=rx, save_path=tmpdir+'/recording')
    SFMdaSortingExtractor.writeSorting(
        sorting=sx, save_path=tmpdir+'/recording/firings_true.mda')

    KiloSort.execute(
        recording_dir=tmpdir+'/recording',
        firings_out=tmpdir+'/firings.mda',
        detect_sign=-1,
        adjacency_radius=50,
        _container='default'
    )
    assert os.path.exists(tmpdir+'/firings.mda')

@pytest.mark.exclude
@pytest.mark.ironclust
def test_ironclust(tmpdir):
    tmpdir = str(tmpdir)

    rx, sx = example_datasets.toy_example1()
    SFMdaRecordingExtractor.writeRecording(
        recording=rx, save_path=tmpdir+'/recording')
    SFMdaSortingExtractor.writeSorting(
        sorting=sx, save_path=tmpdir+'/recording/firings_true.mda')

    IronClust.execute(
        recording_dir=tmpdir+'/recording',
        firings_out=tmpdir+'/firings.mda',
        detect_sign=-1,
        adjacency_radius=50,
        _container='default'
    )
    assert os.path.exists(tmpdir+'/firings.mda')



import uuid
from spikeforest import mdaio
import io
import base64
import vdomr as vd
import os
import numpy as np
import mtlogging
import time
import traceback

source_path=os.path.dirname(os.path.realpath(__file__))

def _mda32_to_base64(X):
    f=io.BytesIO()
    mdaio.writemda32(X,f)
    return base64.b64encode(f.getvalue()).decode('utf-8')

class TemplateWidget(vd.Component):
    def __init__(self,*,template,size=(200,200)):
        vd.Component.__init__(self)

        vd.devel.loadBootstrap()
        vd.devel.loadCss(url='https://stackpath.bootstrapcdn.com/font-awesome/4.7.0/css/font-awesome.min.css')
        vd.devel.loadJavascript(path=source_path+'/mda.js')
        vd.devel.loadJavascript(path=source_path+'/canvaswidget.js')
        vd.devel.loadJavascript(path=source_path+'/templatewidget.js')
        vd.devel.loadJavascript(path=source_path+'/../dist/jquery-3.3.1.min.js')

        self._div_id='TemplateWidget-'+str(uuid.uuid4())
        self._template = template
        self._template_b64=_mda32_to_base64(self._template)
        self._y_scale_factor=None

        self._size=size
    def setYScaleFactor(self, scale_factor):
        self._y_scale_factor=scale_factor
        self.refresh()
    def setSize(self,size):
        if self._size==size:
            return
        self._size=size
        self.refresh()
    def size(self):
        return self._size
    def render(self):
        div=vd.div(id=self._div_id)
        return div
    def postRenderScript(self):
        js="""
        let W=new window.TemplateWidget();
        let X=new window.Mda();
        X.setFromBase64('{template_b64}');
        W.setTemplate(X);
        W.setSize({width},{height});
        W.setYScaleFactor({y_scale_factor});
        $('#{div_id}').empty();
        $('#{div_id}').css({width:'{width}px',height:'{height}px'})
        $('#{div_id}').append(W.element());
        """
        js = js.replace('{template_b64}', self._template_b64)
        js = js.replace('{div_id}', self._div_id)
        js = js.replace('{width}', str(self._size[0]))
        js = js.replace('{height}', str(self._size[1]))
        js = js.replace('{y_scale_factor}', str(self._y_scale_factor or 'null'))
        return js


import mlprocessors as mlpr

import os
import time
import numpy as np
from os.path import join
from subprocess import Popen, PIPE
import shlex
import random
import string
import shutil
from .yasssortingextractor import yassSortingExtractor
from .tools import saveProbeFile

# yass uses negative polarity by default


class YASS(mlpr.Processor):
    NAME = 'YASS'
    VERSION = '0.1.0'
    # used by container to pass the env variables
    ENVIRONMENT_VARIABLES = [
        'NUM_WORKERS', 'MKL_NUM_THREADS', 'NUMEXPR_NUM_THREADS', 'OMP_NUM_THREADS']
    ADDITIONAL_FILES = ['*.yaml']

    # The following container uses python 2
    # CONTAINER = 'sha1://087767605e10761331699dda29519444bbd823f4/02-12-2019/yass.simg'
    
    # this one uses python 3
    CONTAINER = 'sha1://348be6fb09807c774e469c3aeabf4bca867c039f/03-29-2019/yass.simg'
    
    # CONTAINER_SHARE_ID = '69432e9201d0'  # place to look for container

    recording_dir = mlpr.Input('Directory of recording', directory=True)
    channels = mlpr.IntegerListParameter(
        description='List of channels to use.', optional=True, default=[])
    firings_out = mlpr.Output('Output firings file')
    #paramfile_out = mlpr.Output('YASS yaml config file')

    detect_sign = mlpr.IntegerParameter(description='-1, 1, or 0')
    adjacency_radius = mlpr.FloatParameter(
        optional=True, default=100, description='Channel neighborhood adjacency radius corresponding to geom file')
    template_width_ms = mlpr.FloatParameter(
        optional=True, default=3, description='Spike width in milliseconds')
    filter = mlpr.BoolParameter(optional=True, default=True)

    def run(self):
        from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
        import spikeextractors as se

        code = ''.join(random.choice(string.ascii_uppercase)
                       for x in range(10))
        tmpdir = os.environ.get('TEMPDIR', '/tmp')+'/yass-tmp-'+code

        #num_workers = os.environ.get('NUM_WORKERS', 1)
        #print('num_workers: {}'.format(num_workers))
        try:
            recording = SFMdaRecordingExtractor(self.recording_dir)
            if len(self.channels) > 0:
                recording = se.SubRecordingExtractor(
                    parent_recording=recording, channel_ids=self.channels)
            if not os.path.exists(tmpdir):
                os.mkdir(tmpdir)
            sorting, yaml_file = yass_helper(
                recording=recording,
                output_folder=tmpdir,
                probe_file=None,
                file_name=None,
                detect_sign=self.detect_sign,
                adjacency_radius=self.adjacency_radius,
                template_width_ms=self.template_width_ms,
                filter=self.filter)
            SFMdaSortingExtractor.writeSorting(
                sorting=sorting, save_path=self.firings_out)
            #shutil.copyfile(yaml_file, self.paramfile_out)
        except:
            if os.path.exists(tmpdir):
                shutil.rmtree(tmpdir)
            raise
        if not getattr(self, '_keep_temp_files', False):
            shutil.rmtree(tmpdir)


def yass_helper(
        recording,
        output_folder=None,  # Temporary working directory
        probe_file=None,
        file_name=None,
        detect_sign=-1,  # -1 - 1 - 0
        template_width_ms=1,  # yass parameter
        filter=True,
        adjacency_radius=100):

    from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor

    source_dir = os.path.dirname(os.path.realpath(__file__))

    # make output dir
    if output_folder is None:
        output_folder = 'yass'
    else:
        output_folder = join(output_folder, 'yass')
    output_folder = os.path.abspath(output_folder)
    if not os.path.isdir(output_folder):
        os.makedirs(output_folder)

    # save prb file:
    if probe_file is None:
        probe_file = join_abspath_(output_folder, 'probe.npy')
    saveProbeFile(recording, probe_file, format='yass')

    # save binary file
    if file_name is None:
        file_name = 'raw.bin'
    bin_file = join_abspath_(output_folder, file_name)
    # print('bin_file:{}'.format(bin_file))
    writeRecording_(recording=recording, save_path=bin_file,
                    fReversePolarity=(detect_sign > 0), dtype=np.float32, scale_factor=1)
    #print('bin_file exists? {}'.format(os.path.exists(bin_file)))

    # set up yass config file
    print(source_dir)
    with open(join(source_dir, 'config_default.yaml'), 'r') as f:
        yass_config = f.read()

    # get the order
    # root_folder, recordings, geometry, dtype, sampling_rate, n_channels, spatial_radius, spike_size_ms, filter
    n_channels = recording.getNumChannels()
    sampling_rate = recording.getSamplingFrequency()

    # print('sampling_rate={}'.format(sampling_rate))

    yaml_file = join(output_folder, file_name + '.yaml')
    yass_config = yass_config.format(
        output_folder, bin_file, probe_file, 'single', int(sampling_rate), n_channels, adjacency_radius, template_width_ms, filter)
    with open(yaml_file, 'w') as f:
        f.write(yass_config)

    with open(yaml_file) as ff:
        print('YASS CONFIG:')
        print(ff.read())

    print('Running yass...')
    t_start_proc = time.time()

    yass_path = '/usr/local/bin'
    num_cores_str = ''
    # cmd = 'python2 {}/yass {} {} '.format(
    #    yass_path, join(output_folder, file_name+'.yaml'), num_cores_str)
    cmd = 'yass {}'.format(join(output_folder, file_name+'.yaml'))

    retcode = run_command_and_print_output(cmd)
    if retcode != 0:
        raise Exception('yass returned a non-zero exit code')

    # retcode = run_command_and_print_output(cmd_merge)
    # if retcode != 0:
    #    raise Exception('yass merging returned a non-zero exit code')
    processing_time = time.time() - t_start_proc
    print('Elapsed time: ', processing_time)
    sorting = yassSortingExtractor(join_abspath_(output_folder, 'tmp'))

    return sorting, yaml_file


def run_command_and_print_output(command):
    print('RUNNING: '+command)
    with Popen(shlex.split(command), stdout=PIPE, stderr=PIPE) as process:
        while True:
            output_stdout = process.stdout.readline()
            output_stderr = process.stderr.readline()
            if (not output_stdout) and (not output_stderr) and (process.poll() is not None):
                break
            if output_stdout:
                print(output_stdout.decode())
            if output_stderr:
                print(output_stderr.decode())
        rc = process.poll()
        return rc


def join_abspath_(path1, path2):
    path_abs = os.path.abspath(os.path.join(path1, path2))
    return path_abs


def writeRecording_(recording, save_path, dtype=None, transpose=False, fReversePolarity=False, scale_factor=1):
    #save_path = Path(save_path)
    print('writeRecording2: {}'.format(str(save_path)))

    if dtype == None:
        dtype = np.float32
    np_Wav = np.array(recording.getTraces(), dtype=dtype)
    if transpose:
        np_Wav = np.transpose(np_Wav)
    if fReversePolarity:
        np_Wav = np_Wav * -1
    np_Wav = np_Wav * scale_factor
    with open(save_path, 'wb') as f:
        np.ravel(np_Wav, order='F').tofile(f)



from mountaintools import client as mt
from PIL import Image
import json
import pandas as pd
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor

def kb_read_text_file(fname):
    return mt.loadText(path=fname)
    
def kb_read_json_file(fname):
    return mt.loadObject(path=fname)

class SFSortingResult():
    def __init__(self,obj,recording):
        self._obj=obj
        self._recording=recording
    def getObject(self):
        return self._obj
    def recording(self):
        return self._recording
    def sorterName(self):
        return self._obj['sorter']['name']
    def plotNames(self):
        plots=self._obj['summary'].get('plots',dict())
        return list(plots.keys())
    def sorting(self):
        return SFMdaSortingExtractor(firings_file=self._obj['firings'])
    def plot(self,name,format='image'):
        plots=self._obj['summary'].get('plots',dict())
        url=plots[name]
        if format=='url':
            return url
        else:
            path=mt.realizeFile(url)
            if format=='image':
                return Image.open(path)
            elif format=='path':
                return path
            else:
                raise Exception('Invalid format: '+format)
    def comparisonWithTruth(self,*,format='dataframe'):
        A=self._obj['comparison_with_truth']
        if not A:
            return None
        if format=='html':
            return kb_read_text_file(A['html'])
        else:
            B=kb_read_json_file(A['json'])
            if format=='json':
                return B
            elif format=='dataframe':
                return pd.DataFrame(B).transpose()
            else:
                raise Exception('Invalid format: '+format)

class SFRecording():
    def __init__(self,obj,study):
        self._obj=obj
        self._sorting_result_names=[]
        self._sorting_results_by_name=dict()
        self._summary_result=None
        if 'summary' in obj:
            self._summary_result=obj['summary']
        self._study=study
    def getObject(self):
        return self._obj
    def getSummaryObject(self):
        return self._summary_result
    def study(self):
        return self._study
    def name(self):
        return self._obj.get('recording_name',self._obj.get('name'))
    def description(self):
        return self._obj['description']
    def directory(self):
        return self._obj['directory']
    def recordingFileIsLocal(self):
        fname=self.directory()+'/raw.mda'
        fname2=mt.findFile(fname, local_only=True)
        if fname2 and (not _is_url(fname2)):
            return True
        return False
    def realizeRecordingFile(self):
        fname=self.directory()+'/raw.mda'
        return mt.realizeFile(fname)
    def firingsTrueFileIsLocal(self):
        fname=self.directory()+'/firings_true.mda'
        fname2=mt.findFile(fname, local_only=True)
        if fname2 and (not _is_url(fname2)):
            return True
        return False
    def realizeFiringsTrueFile(self):
        fname=self.directory()+'/firings_true.mda'
        return mt.realizeFile(fname)
    def recordingExtractor(self,download=False):
        X=SFMdaRecordingExtractor(dataset_directory=self.directory(),download=download)
        if 'channels' in self._obj:
            if self._obj['channels']:
                X=si.SubRecordingExtractor(parent_recording=X,channel_ids=self._obj['channels'])
        return X
    def sortingTrue(self):
        return SFMdaSortingExtractor(firings_file=self.directory()+'/firings_true.mda')
    def plotNames(self):
        if not self._summary_result:
            return []
        plots=self._summary_result.get('plots',dict())
        return list(plots.keys())
    def plot(self,name,format='image'):
        plots=self._summary_result.get('plots',dict())
        url=plots[name]
        if format=='url':
            return url
        else:
            path=mt.realizeFile(url)
            if format=='image':
                return Image.open(path)
            elif format=='path':
                return path
            else:
                raise Exception('Invalid format: '+format)
    def trueUnitsInfo(self,format='dataframe'):
        if not self._summary_result:
            return None
        B=kb_read_json_file(self._summary_result['true_units_info'])
        if format=='json':
            return B
        elif format=='dataframe':
            return pd.DataFrame(B)
        else:
            raise Exception('Invalid format: '+format)
    def setSummaryResult(self,obj):
        self._summary_result=obj
    def addSortingResult(self,obj):
        sorter_name=obj['sorter']['name']
        if sorter_name in self._sorting_results_by_name:
            print ('Sorting result already in recording: {}'.format(sorter_name))
        else:
            R=SFSortingResult(obj,self)
            self._sorting_result_names.append(sorter_name)
            self._sorting_results_by_name[sorter_name]=R
    def sortingResultNames(self):
        return self._sorting_result_names
    def sortingResult(self,name):
        return self._sorting_results_by_name.get(name,None)

class SFStudy():
    def __init__(self,obj):
        self._obj=obj
        self._recordings_by_name=dict()
        self._recording_names=[]
    def getObject(self):
        return self._obj
    def name(self):
        return self._obj.get('recording_name',self._obj.get('name'))
    def description(self):
        return self._obj['description']
    def addRecording(self,obj):
        name=obj.get('recording_name',obj.get('name'))
        if name in self._recordings_by_name:
            print ('Recording already in study: '+name)
        else:
            self._recording_names.append(name)
            D=SFRecording(obj,self)
            self._recordings_by_name[name]=D
    def recordingNames(self):
        return self._recording_names
    def recording(self,name):
        return self._recordings_by_name.get(name,None)
        

class SFData():
    def __init__(self):
        self._studies_by_name=dict()
        self._study_names=[]
    def loadStudy(self,study):
        name=study['name']
        if name in self._studies_by_name:
            print ('Study already loaded: '+name)
        else:
            self._study_names.append(study['name'])
            S=SFStudy(study)
            self._studies_by_name[name]=S
    def loadStudies(self,studies):
        for study in studies:
            self.loadStudy(study)
    def loadRecording(self,recording):
        study=recording.get('study_name',recording.get('study_name',recording.get('study')))
        self._studies_by_name[study].addRecording(recording)
    def loadRecordings2(self,recordings):
        for recording in recordings:
            self.loadRecording(recording)
    def loadRecordings(self,*,key=None,verbose=False):
        # old
        if key is None:
            key=dict(name='spikeforest_studies_processed')
        obj=mt.loadObject(key=key)
        studies=obj['studies']
        for study in studies:
            name=study['name']
            if name in self._studies_by_name:
                print ('Study already loaded: '+name)
            else:
                self._study_names.append(study['name'])
                S=SFStudy(study)
                self._studies_by_name[name]=S
        recordings=obj['recordings']
        print('recordings ===================================================================')
        print(recordings)
        for ds in recordings:
            study=ds.get('study_name',ds.get('study'))
            self._studies_by_name[study].addRecording(ds)
        if verbose:
            print ('Loaded {} recordings'.format(len(recordings)))
    def loadSortingResults(self,sorting_results):
        for result in sorting_results:
            self.loadSortingResult(result)
    def loadSortingResult(self,X):
        study_name=X['recording'].get('study_name',X['recording'].get('study'))
        recording_name=X['recording'].get('recording_name',X['recording'].get('name'))
        # sorter_name=X['sorter']['name']
        S=self.study(study_name)
        if S:
            D=S.recording(recording_name)
            if D:
                D.addSortingResult(X)
            else:
                print ('Warning: recording not found: '+recording_name)
        else:
            print ('Warning: study not found: '+study_name)

    def studyNames(self):
        return self._study_names
    def study(self,name):
        return self._studies_by_name.get(name,None)

def _is_url(path):
    return (path.startswith('http://') or path.startswith('https://'))



import mlprocessors as mlpr
from mountaintools import client as mt
import os
from copy import deepcopy
import mtlogging

@mtlogging.log()
def summarize_sortings(sortings,compute_resource,label=None):
    print('')
    print('>>>>>> {}'.format(label or 'summarize sortings'))
    
    print('Gathering summarized sortings...')
    summarized_sortings=[]
    for sorting in sortings:
        summary=dict()
        summary['plots']=dict()

        sorting2=deepcopy(sorting)
        sorting2['summary']=summary
        summarized_sortings.append(sorting2)

    return summarized_sortings



import mlprocessors as mlpr
import os
import time
import numpy as np
from os.path import join
from subprocess import Popen, PIPE
import shlex
import random
import string
import shutil
from spikeforest import mdaio
import spikeextractors as se
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
import sys
import shlex
#import h5py


class KiloSort(mlpr.Processor):
    NAME = 'KiloSort'
    VERSION = '0.2.0'  # wrapper VERSION
    ADDITIONAL_FILES = ['*.m']
    ENVIRONMENT_VARIABLES = [
        'NUM_WORKERS', 'MKL_NUM_THREADS', 'NUMEXPR_NUM_THREADS', 'OMP_NUM_THREADS']
    CONTAINER = None
    CONTAINER_SHARE_ID = None

    recording_dir = mlpr.Input('Directory of recording', directory=True)
    channels = mlpr.IntegerListParameter(
        description='List of channels to use.', optional=True, default=[])
    firings_out = mlpr.Output('Output firings file')

    detect_sign = mlpr.IntegerParameter(
        'Use -1 or 1, depending on the sign of the spikes in the recording')
    adjacency_radius = mlpr.FloatParameter(
        'Use -1 to include all channels in every neighborhood')
    detect_threshold = mlpr.FloatParameter(
        optional=True, default=3, description='')
    # prm_template_name=mlpr.StringParameter(optional=False,description='TODO')
    freq_min = mlpr.FloatParameter(
        optional=True, default=300, description='Use 0 for no bandpass filtering')
    freq_max = mlpr.FloatParameter(
        optional=True, default=6000, description='Use 0 for no bandpass filtering')
    merge_thresh = mlpr.FloatParameter(
        optional=True, default=0.98, description='TODO')
    pc_per_chan = mlpr.IntegerParameter(
        optional=True, default=3, description='TODO')

    def run(self):
        keep_temp_files = False
        code = ''.join(random.choice(string.ascii_uppercase)
                       for x in range(10))
        tmpdir = os.environ.get('TEMPDIR', '/tmp')+'/kilosort-tmp-'+code

        try:
            recording = SFMdaRecordingExtractor(self.recording_dir)
            if len(self.channels) > 0:
                recording = se.SubRecordingExtractor(
                    parent_recording=recording, channel_ids=self.channels)
            if not os.path.exists(tmpdir):
                os.mkdir(tmpdir)
            sorting = kilosort_helper(
                recording=recording,
                tmpdir=tmpdir,
                detect_sign=self.detect_sign,
                adjacency_radius=self.adjacency_radius,
                detect_threshold=self.detect_threshold,
                merge_thresh=self.merge_thresh,
                freq_min=self.freq_min,
                freq_max=self.freq_max,
                pc_per_chan=self.pc_per_chan
            )
            SFMdaSortingExtractor.writeSorting(
                sorting=sorting, save_path=self.firings_out)
        except:
            if os.path.exists(tmpdir):
                if not keep_temp_files:
                    shutil.rmtree(tmpdir)
            raise
        if not keep_temp_files:
            shutil.rmtree(tmpdir)


def kilosort_helper(*,
                    recording,  # Recording object
                    tmpdir,  # Temporary working directory
                    detect_sign=-1,  # Polarity of the spikes, -1, 0, or 1
                    adjacency_radius=-1,  # Channel neighborhood adjacency radius corresponding to geom file
                    detect_threshold=5,  # Threshold for detection
                    merge_thresh=.98,  # Cluster merging threhold 0..1
                    freq_min=300,  # Lower frequency limit for band-pass filter
                    freq_max=6000,  # Upper frequency limit for band-pass filter
                    pc_per_chan=3,  # Number of pc per channel
                    KILOSORT_PATH=None,  # github kilosort
                    IRONCLUST_PATH=None  # github ironclust
                    ):
    if KILOSORT_PATH is None:
        KILOSORT_PATH = os.getenv('KILOSORT_PATH', None)
    if not KILOSORT_PATH:
        raise Exception(
            'You must either set the KILOSORT_PATH environment variable, or pass the KILOSORT_PATH parameter')

    if IRONCLUST_PATH is None:
        IRONCLUST_PATH = os.getenv('IRONCLUST_PATH', None)
    if not IRONCLUST_PATH:
        raise Exception(
            'You must either set the IRONCLUST_PATH environment variable, or pass the IRONCLUST_PATH parameter')

    source_dir = os.path.dirname(os.path.realpath(__file__))

    dataset_dir = tmpdir+'/kilosort_dataset'
    # Generate three files in the dataset directory: raw.mda, geom.csv, params.json
    SFMdaRecordingExtractor.writeRecording(
        recording=recording, save_path=dataset_dir)

    samplerate = recording.getSamplingFrequency()

    print('Reading timeseries header...')
    HH = mdaio.readmda_header(dataset_dir+'/raw.mda')
    num_channels = HH.dims[0]
    num_timepoints = HH.dims[1]
    duration_minutes = num_timepoints/samplerate/60
    print('Num. channels = {}, Num. timepoints = {}, duration = {} minutes'.format(
        num_channels, num_timepoints, duration_minutes))

    print('Creating argfile.txt file...')
    txt = ''
    txt += 'samplerate={}\n'.format(samplerate)
    txt += 'detect_sign={}\n'.format(detect_sign)
    txt += 'adjacency_radius={}\n'.format(adjacency_radius)
    txt += 'detect_threshold={}\n'.format(detect_threshold)
    txt += 'merge_thresh={}\n'.format(merge_thresh)
    txt += 'freq_min={}\n'.format(freq_min)
    txt += 'freq_max={}\n'.format(freq_max)
    txt += 'pc_per_chan={}\n'.format(pc_per_chan)
    _write_text_file(dataset_dir+'/argfile.txt', txt)

    print('Running kilosort in {tmpdir}...'.format(tmpdir=tmpdir))
    cmd = '''
addpath('{source_dir}');
try
    p_kilosort('{ksort}', '{iclust}', '{tmpdir}', '{raw}', '{geom}', '{firings}', '{arg}');
catch
    quit(1);
end
quit(0);
        '''
    cmd = cmd.format(source_dir=source_dir, ksort=KILOSORT_PATH, iclust=IRONCLUST_PATH, \
            tmpdir=tmpdir, raw=dataset_dir+'/raw.mda', geom=dataset_dir+'/geom.csv', \
            firings=tmpdir+'/firings.mda', arg=dataset_dir+'/argfile.txt')
    matlab_cmd = mlpr.ShellScript(cmd,script_path=tmpdir+'/run_kilosort.m',keep_temp_files=True)
    matlab_cmd.write();
    shell_cmd = '''
        #!/bin/bash
        cd {tmpdir}
        echo '=====================' `date` '=====================' >> run_kilosort.log 
        matlab -nosplash -nodisplay -r run_kilosort &>> run_kilosort.log
    '''.format(tmpdir=tmpdir)
    shell_cmd = mlpr.ShellScript(shell_cmd, script_path=tmpdir+'/run_kilosort.sh', keep_temp_files=True)
    shell_cmd.write(tmpdir+'/run_kilosort.sh')
    shell_cmd.start()
    retcode = shell_cmd.wait()

    if retcode != 0:
        raise Exception('kilosort returned a non-zero exit code')

    # parse output
    result_fname = tmpdir+'/firings.mda'
    if not os.path.exists(result_fname):
        raise Exception('Result file does not exist: ' + result_fname)

    firings = mdaio.readmda(result_fname)
    sorting = se.NumpySortingExtractor()
    sorting.setTimesLabels(firings[1, :], firings[2, :])
    return sorting


def _read_text_file(fname):
    with open(fname) as f:
        return f.read()


def _write_text_file(fname, str):
    with open(fname, 'w') as f:
        f.write(str)




import numpy as np
import json
import mlprocessors as mlpr
from mountaintools import client as mt
import spikeextractors as si
from spikeforest import spikewidgets as sw
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor

_CONTAINER='sha1://1ad2478736ad188ab5050289ffb1d2c29d1ba750/03-29-2019/spikeforest_basic.simg'

def write_json_file(fname,obj):
  with open(fname, 'w') as f:
    json.dump(obj, f)

def get_random_spike_waveforms(*,recording,sorting,unit,snippet_len,max_num,channels=None):
    st=sorting.getUnitSpikeTrain(unit_id=unit)
    num_events=len(st)
    if num_events>max_num:
        event_indices=np.random.choice(range(num_events),size=max_num,replace=False)
    else:
        event_indices=range(num_events)

    spikes=recording.getSnippets(reference_frames=st[event_indices].astype(int),snippet_len=snippet_len,channel_ids=channels)
    if len(spikes)>0:
      spikes=np.dstack(tuple(spikes))
    else:
      spikes=np.zeros((recording.getNumChannels(),snippet_len,0))
    return spikes

def compute_unit_templates(*,recording,sorting,unit_ids,snippet_len=50,max_num=100,channels=None):
    ret=[]
    for unit in unit_ids:
        # print('Unit {} of {}'.format(unit,len(unit_ids)))
        waveforms=get_random_spike_waveforms(recording=recording,sorting=sorting,unit=unit,snippet_len=snippet_len,max_num=max_num,channels=None)
        template=np.median(waveforms,axis=2)
        ret.append(template)
    return ret

def compute_template_snr(template,channel_noise_levels):
    channel_snrs=[]
    for ch in range(template.shape[0]):
        #channel_snrs.append((np.max(template[ch,:])-np.min(template[ch,:]))/channel_noise_levels[ch])
        channel_snrs.append((np.max(np.abs(template[ch,:])))/channel_noise_levels[ch])
    return float(np.max(channel_snrs))
    
def compute_channel_noise_levels(recording):
    channel_ids=recording.getChannelIds()
    # M=len(channel_ids)
    samplerate=int(recording.getSamplingFrequency())
    X=recording.getTraces(start_frame=samplerate*1,end_frame=samplerate*2)
    ret=[]
    for ii in range(len(channel_ids)):
        #noise_level=np.std(X[ii,:])
        noise_level=np.median(np.abs(X[ii,:]))/0.6745 # median absolute deviation (MAD)
        ret.append(noise_level)
    return ret

class ComputeUnitsInfo(mlpr.Processor):
  NAME='ComputeUnitsInfo'
  VERSION='0.1.8'
  CONTAINER=_CONTAINER
  recording_dir=mlpr.Input(directory=True,description='Recording directory')
  channel_ids=mlpr.IntegerListParameter(description='List of channels to use.',optional=True,default=[])
  unit_ids=mlpr.IntegerListParameter(description='List of units to use.',optional=True,default=[])
  firings=mlpr.Input(description='Firings file')
  json_out=mlpr.Output('The info as a .json file')
  
  def run(self):
    R0=SFMdaRecordingExtractor(dataset_directory=self.recording_dir,download=True)
    sorting=SFMdaSortingExtractor(firings_file=self.firings)
    ret = compute_units_info(recording=R0, sorting=sorting, channel_ids=self.channel_ids, unit_ids=self.unit_ids)
    write_json_file(self.json_out,ret)
    

def compute_units_info(*, recording, sorting, channel_ids=[], unit_ids=[]):
  if (channel_ids) and (len(channel_ids)>0):
      recording=si.SubRecordingExtractor(parent_recording=recording, channel_ids=channel_ids)

  # load into memory
  print('Loading recording into RAM...')
  recording=si.NumpyRecordingExtractor(timeseries=recording.getTraces(), samplerate=recording.getSamplingFrequency())

  # do filtering
  print('Filtering...')
  recording=sw.lazyfilters.bandpass_filter(recording=recording,freq_min=300,freq_max=6000)
  recording=si.NumpyRecordingExtractor(timeseries=recording.getTraces(), samplerate=recording.getSamplingFrequency())

  
  if (not unit_ids) or (len(unit_ids)==0):
    unit_ids=sorting.getUnitIds()

  print('Computing channel noise levels...')
  channel_noise_levels=compute_channel_noise_levels(recording=recording)

  # No longer use subset to compute the templates
  print('Computing unit templates...')
  templates=compute_unit_templates(recording=recording, sorting=sorting, unit_ids=unit_ids, max_num=100)

  print(recording.getChannelIds())

  ret=[]
  for i,unit_id in enumerate(unit_ids):
    print('Unit {} of {} (id={})'.format(i+1, len(unit_ids), unit_id))
    template=templates[i]
    max_p2p_amps_on_channels=np.max(template,axis=1)-np.min(template,axis=1)
    peak_channel_index=np.argmax(max_p2p_amps_on_channels)
    peak_channel=recording.getChannelIds()[peak_channel_index]
    peak_signal = np.max(np.abs(template[peak_channel_index,:]))
    info0=dict()
    info0['unit_id']=int(unit_id)
    info0['snr']=peak_signal/channel_noise_levels[peak_channel_index]
    info0['peak_channel']=int(recording.getChannelIds()[peak_channel])
    train=sorting.getUnitSpikeTrain(unit_id=unit_id)
    info0['num_events']=int(len(train))
    info0['firing_rate']=float(len(train)/(recording.getNumFrames()/recording.getSamplingFrequency()))
    ret.append(info0)
  return ret
  
  
## return format can be 'json' or 'filename'
def compute_units_info_b(*,recording_dir,firings,channel_ids=[],unit_ids=[],return_format='json'):
    out=ComputeUnitsInfo.execute(
      recording_dir=recording_dir,
      firings=firings,
      unit_ids=unit_ids,
      channel_ids=channel_ids,
      json_out={'ext':'.json'},
      _container='default'
    ).outputs
    fname=out['json_out']
    if return_format=='filename':
      return fname
    else:
      fname=mt.realizeFile(path=fname)
      with open(fname) as f:
        return json.load(f)

def select_units_on_channels(recording_dir,firings,channels):
  info=compute_units_info_b(recording_dir=recording_dir,firings=firings)
  units=[]
  for info0 in info:
    if info0['peak_channel'] in channels:
      units.append(info0['unit_id'])
  return units



from matplotlib import pyplot as plt
import numpy as np

class ElectrodeGeometryWidget:
  def __init__(self, *, recording):
    self._recording = recording

  def plot(self,width=1.5,height=1.5):
    self._do_plot(width=width,height=height)

  def _get_geom(self):
    RX=self._recording
    return np.stack([RX.getChannelProperty(channel_id=ch,property_name='location') for ch in RX.getChannelIds()])
    
  def _do_plot(self,width,height):
    R=self._recording
    geom=self._get_geom()
    
    fig = plt.figure(figsize=(width,height))
    ax = fig.add_axes([0, 0, 1, 1])
    ax.axis('off')
    
    x=geom[:,0]
    y=geom[:,1]
    xmin=np.min(x); xmax=np.max(x)
    ymin=np.min(y); ymax=np.max(y)
    
    marker_size=width*fig.dpi/6
    margin=np.maximum(xmax-xmin,ymax-ymin)*0.2

    plt.scatter(x,y,marker='o', s=int(marker_size**2))
    plt.axis('equal')
    plt.xticks([])
    plt.yticks([])
    plt.xlim(xmin-margin,xmax+margin)
    plt.ylim(ymin-margin,ymax+margin)
    #plt.show()


# pylint: disable=no-member

import mlprocessors as mlpr
import spikeextractors as se
import h5py
import numpy as np
import spikeextractors as se
from mountaintools import client as mt
import mtlogging

class EfficientAccessRecordingExtractor(se.RecordingExtractor):
    def __init__(self, *, path=None, recording=None, _dest_path=None):
        se.RecordingExtractor.__init__(self)
        if path is not None:
            if recording is not None:
                raise Exception('Cannot pass both path and recording to EfficientAccessRecordingExtractor.')
            self._path = path
        elif recording is not None:
            if not hasattr(recording, 'hash'):
                try:
                    if not hasattr(recording, 'hash'):
                        print('''
                        Warning: Recording does not have the hash attribute.
                        Using sampling method to compute a hash.''')
                        setattr(recording, 'hash', _hash(recording))
                except:
                    raise Exception('Recording class does not support the sampling hash.')
            path0 = CreateEfficientAccessRecordingFile.execute(
                recording=recording,
                hdf5_out=dict(ext='.hdf5', dest_path=_dest_path)
            ).outputs['hdf5_out']
            self._path = mt.realizeFile(path=path0)
        else:
            raise Exception('Missing argument: path or recording')
        
        with h5py.File(self._path, "r") as f:
            self._num_segments = int(np.array(f.get('num_segments'))[0])
            self._segment_size = int(np.array(f.get('segment_size'))[0])
            self._num_channels = int(np.array(f.get('num_channels'))[0])
            self._channel_ids = np.array(f.get('channel_ids')).tolist()
            self._num_timepoints = int(np.array(f.get('num_timepoints'))[0])
            self._samplerate = np.array(f.get('samplerate'))[0]
            self._recording_hash = np.array(f.get('recording_hash'))[0].decode()
            assert type(self._recording_hash)==str
            geom = np.array(f.get('geom'))
            channel_locations = [geom[m, :].ravel() for m in range(self._num_channels)]
            self.setChannelLocations(channel_ids=self._channel_ids, locations=channel_locations)
    
    def hash(self):
        return self._recording_hash

    def path(self):
        return self._path
        
    def getChannelIds(self):
        return self._channel_ids

    def getNumFrames(self):
        return self._num_timepoints

    def getSamplingFrequency(self):
        return self._samplerate

    @mtlogging.log(name='EfficientAccessRecordingExtractor:getTraces')
    def getTraces(self, channel_ids=None, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self.getNumFrames()
        if channel_ids is None:
            channel_ids = self.getChannelIds()
        
        t1 = int(start_frame)
        t2 = int(end_frame)
        c1 = int(t1/self._segment_size)
        c2 = int((t2-1)/self._segment_size)
        ret = np.zeros((len(channel_ids), t2-t1))
        with h5py.File(self._path,"r") as f:
            for cc in range(c1, c2+1):
                if cc==c1:
                    t1a=t1
                else:
                    t1a=self._segment_size*cc
                if cc==c2:
                    t2a=t2
                else:
                    t2a=self._segment_size*(cc+1)
                for ii, m in enumerate(channel_ids):
                    str0 = 'part-{}-{}'.format(m, cc)
                    offset = self._segment_size*cc
                    ret[ii, t1a-t1:t2a-t1] = f[str0][t1a-offset:t2a-offset]
        return ret
    
    @staticmethod
    def writeRecording(recording, save_path):
        EfficientAccessRecordingExtractor(recording=recording, _dest_path=save_path)

class CreateEfficientAccessRecordingFile(mlpr.Processor):
    NAME = 'CreateEfficientAccessRecordingFile'
    VERSION = '0.1.2'
    recording = mlpr.Input()
    segment_size = mlpr.IntegerParameter(optional=True, default=1000000)
    hdf5_out = mlpr.Output()
    
    def run(self):
        recording = self.recording
        segment_size = self.segment_size
        channel_ids = recording.getChannelIds()
        samplerate = recording.getSamplingFrequency()
        M = len(channel_ids) # number of channels
        N = recording.getNumFrames() # Number of timepoints
        num_segments = int(np.ceil(N/segment_size))
        
        channel_locations = recording.getChannelLocations(channel_ids = channel_ids)
        nd = len(channel_locations[0])
        geom = np.zeros((M, nd))
        for m in range(M):
            geom[m, :] = channel_locations[m]
            
        with h5py.File(self.hdf5_out, "w") as f:
            f.create_dataset('segment_size', data=[segment_size])
            f.create_dataset('num_segments', data=[num_segments])
            f.create_dataset('num_channels', data=[M])
            f.create_dataset('channel_ids', data=np.array(channel_ids))
            f.create_dataset('num_timepoints', data=[N])
            f.create_dataset('samplerate', data=[samplerate])
            f.create_dataset('geom', data=geom)
            if callable(recording.hash):
                hash0 = recording.hash()
            else:
                hash0 = recording.hash
            f.create_dataset('recording_hash', data=np.array([hash0.encode()]))
            for j in range(num_segments):
                segment = np.zeros((M,segment_size),dtype=float) # fix dtype here
                t1 = int(j*segment_size) # first timepoint of the segment
                t2 = int(np.minimum(N,(t1+segment_size))) # last timepoint of segment (+1)
                s1 = int(np.maximum(0,t1)) # first timepoint
                s2 = int(np.minimum(N,t2)) # last timepoint (+1)

                # determine aa so that t1-s1+aa = 0
                # so, aa = -(t1-s1)
                aa = -(t1-s1)
                segment[:,aa:aa+s2-s1] = recording.getTraces(start_frame=s1, end_frame=s2) # Read the segment

                for ii, ch in enumerate(channel_ids):
                    f.create_dataset('part-{}-{}'.format(ch, j), data=segment[ii,:].ravel())

def _hash(self):
    from mountaintools import client as mt
    obj = {
        'channels': tuple(self.getChannelIds()),
        'frames': self.getNumFrames(),
        'data': samplehash(self)
    }
    return mt.sha1OfObject(obj)
    
def samplehash(self):
    rng = np.random.RandomState(37)
    n_samples = min(self.getNumFrames()//1000, 100)
    inds = rng.randint(low=0, high=self.getNumFrames(), size=n_samples)
    h = 0
    for i in inds:
        t = self.getTraces(start_frame=i, end_frame=i+100)
        h = hash((hash(bytes(t)),hash(h)))
    return h




from .currentstateview import CurrentStateView
from .recordingtableview import RecordingTableView

def get_spikeforest_view_launchers(context):
    return [
        dict(
            name='recording-table',
            label='Recording table',
            view_class=RecordingTableView,
            context=context,
            opts=dict(),
            enabled=True
        ),
        dict(
            name='current-state',
            label='Current state',
            view_class=CurrentStateView,
            context=context,
            opts=dict(),
            enabled=True
        ),
        dict(
            name='test-name',
            label='Testing',
            view_class=CurrentStateView,
            context=context,
            opts=dict(),
            enabled=(context.currentRecordingId() is not None)
        )
    ]




from mountaintools import client as mt
import random
import string
import datetime

def clearBatch(*,batch_name, test_one=False):
  batch=mt.loadObject(key=dict(batch_name=batch_name))
  jobs=batch['jobs']

  if test_one and (len(jobs)>0):
    jobs=[jobs[0]]

  setBatchStatus(
    batch_name=batch_name,
    status='clearing_batch'
  )
  _clear_job_results(jobs=jobs,incomplete_only=False)
  setBatchStatus(
    batch_name=batch_name,
    status='finished_clearing_batch'
  )


def prepareBatch(*,batch_name, test_one=False):
  batch=mt.loadObject(key=dict(batch_name=batch_name))
  jobs=batch['jobs']

  if test_one and (len(jobs)>0):
    jobs=[jobs[0]]

  setBatchStatus(
    batch_name=batch_name,
    status='preparing_batch'
  )
  _clear_job_results(jobs=jobs,incomplete_only=True)

  setBatchStatus(
    batch_name=batch_name,
    status='downloading_recordings'
  )
  _download_recordings(jobs=jobs)

  setBatchStatus(
    batch_name=batch_name,
    status='finished_preparing_batch'
  )

def runBatch(*,batch_name, test_one=False):
  print('Loading batch object...')
  batch=mt.loadObject(key=dict(batch_name=batch_name))
  jobs=batch['jobs']

  if test_one and (len(jobs)>0):
    jobs=[jobs[0]]

  print('Running batch with {} jobs...'.format(len(jobs)))
  for job in jobs:
    _run_job(job)

def setBatchStatus(*,batch_name,status):
  mt.setValue(key=dict(name='spikeforest_batch_status',batch_name=batch_name),value=status)

def _do_sort_recording(job):
  try:
    from .sort_recording import sort_recording
  except:
    print('Problem importing sort_recording. You probably need to install one or more python packages.')
    raise
  return sort_recording(sorter=job['sorter'],recording=job['recording'])

def _do_summarize_recording(job):
  try:
    from .summarize_recording import summarize_recording
  except:
    print('Problem importing summarize_recording. You probably need to install one or more python packages.')
    raise
  return summarize_recording(recording=job['recording'])

def _do_run_job(job):
  if job['command']=='sort_recording':
    return _do_sort_recording(job)
  elif job['command']=='summarize_recording':
    return _do_summarize_recording(job)
  else:
    return dict(error='Invalid job command: '+job['command'])

def _set_job_status(job,status):
  mt.saveObject(key=dict(name='job_status',job=job),object=status)

def _run_job(job):
  val=mt.getValue(key=job)
  if val:
    return
  code=''.join(random.choice(string.ascii_uppercase) for x in range(10))
  if not mt.setValue(key=job,value='in-process-'+code,overwrite=False):
    return
  status=dict(
    time_started=_make_timestamp(),
    status='running'
  )
  _set_job_status(job,status)

  print('Running job: '+job['label'])
  try:
    result=_do_run_job(job)
  except:
    status['time_finished']=_make_timestamp()
    status['status']='error'
    status['error']='Exception in _do_run_job'
    val=mt.getValue(key=job)
    if val=='in-process-'+code:
      _set_job_status(job,status)  
    raise

  val=mt.getValue(key=job)
  if val!='in-process-'+code:
    print('Not saving result because in-process code does not match {} <> {}.'.format(val,'in-process-'+code))
    return

  status['time_finished']=_make_timestamp()
  status['result']=result
  if 'error' in result:
    print('Error running job: '+result['error'])
    status['status']='error'
    status['error']=result['error']
    _set_job_status(job,status)
    mt.setValue(key=job,value='error-'+code)
    return
  status['status']='finished'
  mt.saveObject(key=job,object=result) # Not needed in future, because we should instead use the status object

def assembleBatchResults(*,batch_name):
  batch=mt.loadObject(key=dict(batch_name=batch_name))
  jobs=batch['jobs']

  print('Assembling results for batch {} with {} jobs'.format(batch_name,len(jobs)))
  job_results=[]
  for job in jobs:
    print('ASSEMBLING: '+job['label'])
    result=mt.loadObject(key=job)
    if not result:
      raise Exception('Unable to load object for job: '+job['label'])
    job_results.append(dict(
        job=job,
        result=result
    ))
  print('Saving results...')
  mt.saveObject(key=dict(name='job_results',batch_name=batch_name),object=dict(job_results=job_results))
  print('Done.')

def _clear_job_results(*,jobs,incomplete_only=True):
  for job in jobs:
    val=mt.getValue(key=job)
    if val:
      if (not incomplete_only) or (val.startswith('in-process')) or (val.startswith('error')):
        print('Clearing job: '+job['label'])
        mt.setValue(key=job,value=None)

def _download_recordings(*,jobs):
  for ii,job in enumerate(jobs):
    val=mt.getValue(key=job)
    if not val:
      if 'recording' in job:
        if 'directory' in job['recording']:
          dsdir=job['recording']['directory']
          fname=dsdir+'/raw.mda'
          print('REALIZING FILE: '+fname)
          mt.realizeFile(path=fname)

def _make_timestamp():
  return str(datetime.datetime.now())


from mountaintools import client as mt
import spikeextractors as si
import mlprocessors as mlpr
import os
import shutil
import random
import string
import multiprocessing
import mtlogging
#from . import sorters as sorters

from spikesorters import MountainSort4, SpykingCircus, YASS, IronClust, KiloSort, KiloSort2, MountainSort4TestError
        
Processors=dict(
    MountainSort4=(MountainSort4, 'default'),
    IronClust=(IronClust,None),
    SpykingCircus=(SpykingCircus,'default'),
    KiloSort=(KiloSort,None),
    KiloSort2=(KiloSort2,None),
    Yass=(YASS,'default'),
    MountainSort4TestError=(MountainSort4TestError, 'default')
)

@mtlogging.log()
def sort_recordings(*,sorter,recordings,compute_resource=None,num_workers=None,disable_container=False, job_timeout=60*20, label=None):
    print('')
    print('>>>>>> {}'.format(label or 'sort recordings'))
    sorting_params=sorter['params']
    processor_name=sorter['processor_name']
    if processor_name in Processors:
        SS=Processors[processor_name][0]
        SS_container=Processors[processor_name][1]
        if disable_container:
            SS_container=None
    else:
        raise Exception('No such sorter: '+processor_name)

    if SS_container:
        if SS_container=='default':
            SS_container=SS.CONTAINER
        print('Locating container: '+SS_container)
        if not mt.findFile(path=SS_container):
            raise Exception('Unable to realize container: '+SS_container)
        
    print('>>>>>>>>>>> Sorting recordings using {}'.format(processor_name))

    # pool = multiprocessing.Pool(20)
    # sorting_jobs=pool.map(_create_sorting_job_for_recording_helper, [dict(recording=recording, sorter=sorter, job_timeout=job_timeout) for recording in recordings])
    # pool.close()
    # pool.join()

    sorting_params=sorter['params']
    processor_name=sorter['processor_name']

    sorting_jobs=SS.createJobs([
        dict(
            _container=SS_container,
            _timeout=job_timeout,
            _label='Sort recording {}/{} using {}'.format(recording.get('study', ''), recording.get('name', ''), sorter.get('name', '')),
            _additional_files_to_realize=[recording['directory']+'/raw.mda'],
            recording_dir=recording['directory'],
            channels=recording.get('channels',[]),
            firings_out=dict(ext='.mda',upload=True),
            **sorting_params
        )
        for recording in recordings
    ])

    label=label or 'Sort recordings using {}'.format(processor_name)
    sorting_job_results = mlpr.executeBatch(jobs=sorting_jobs,label=label,compute_resource=compute_resource,num_workers=num_workers)
    
    print('Gathering sorting results...')
    sorting_results = [
        dict(
            recording=recording,
            sorter=sorter,
            firings_true=recording['directory']+'/firings_true.mda',
            processor_name=SS.NAME,
            processor_version=SS.VERSION,
            execution_stats=sorting_job_results[ii].runtime_info,
            console_out=sorting_job_results[ii].console_out,
            container=SS_container,
            firings=(sorting_job_results[ii].outputs or dict()).get('firings_out', None)
        )
        for ii, recording in enumerate(recordings)
    ]

    return sorting_results
    

    



import os
import sys

# def append_to_path(dir0): # A convenience function
#     if dir0 not in sys.path:
#         sys.path.append(dir0)
# append_to_path(os.getcwd()+'/..')
from spikesorters import KiloSort2
from spikeforest import example_datasets
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
import tempfile
import shutil
from mountaintools import client as mt
import pytest


def setup_module(module):
    # Use this so we can download containers as needed
    mt.configDownloadFrom(['spikeforest.spikeforest2'])

    # Check if singularity is installed
    retval = os.system('singularity --version')
    assert retval == 0, 'Singularity is not installed'


def teardown_module(module):
    pass


def setup_function(function):
    pass


def teardown_function(function):
    pass


@pytest.mark.kilosort2
def test_kilosort2(tmpdir):
    tmpdir = str(tmpdir)

    #rx, sx = example_datasets.toy_example1()
    rx = SFMdaRecordingExtractor('/mnt/home/jjun/ceph/groundtruth/hybrid_drift/rec_64c_1200s_11')
    sx = SFMdaSortingExtractor('/mnt/home/jjun/ceph/groundtruth/hybrid_drift/rec_64c_1200s_11/firings_true.mda')

    SFMdaRecordingExtractor.writeRecording(
        recording=rx, save_path=tmpdir+'/recording')
    SFMdaSortingExtractor.writeSorting(
        sorting=sx, save_path=tmpdir+'/recording/firings_true.mda')

    KiloSort2.execute(
        recording_dir=tmpdir+'/recording',
        firings_out=tmpdir+'/firings.mda',
        detect_sign=-1,
        adjacency_radius=50,
        _container='default',
        _keep_temp_files=True,
        _force_run=True
    )
    assert os.path.exists(tmpdir+'/firings.mda')



import mlprocessors as mlpr
import os
import time
import numpy as np
from os.path import join
from subprocess import Popen, PIPE
import shlex
import random
import string
import shutil
from spikeforest import mdaio
import spikeextractors as se
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
from mountaintools import client as mt
import json


class IronClust(mlpr.Processor):
    NAME = 'IronClust'
    VERSION = '0.2.4'
    ENVIRONMENT_VARIABLES = [
        'NUM_WORKERS', 'MKL_NUM_THREADS', 'NUMEXPR_NUM_THREADS', 'OMP_NUM_THREADS']
    ADDITIONAL_FILES = ['*.m']
    CONTAINER = None
    CONTAINER_SHARE_ID = None

    recording_dir = mlpr.Input('Directory of recording', directory=True)
    channels = mlpr.IntegerListParameter(
        description='List of channels to use.', optional=True, default=[])
    firings_out = mlpr.Output('Output firings file')

    detect_sign = mlpr.IntegerParameter(
        'Use -1, 0, or 1, depending on the sign of the spikes in the recording')
    adjacency_radius = mlpr.FloatParameter(
        optional=True, default=75, description='Use -1 to include all channels in every neighborhood')
    detect_threshold = mlpr.FloatParameter(
        optional=True, default=4.5, description='detection threshold')
    prm_template_name = mlpr.StringParameter(
        optional=True, description='.prm template file name')
    freq_min = mlpr.FloatParameter(
        optional=True, default=300, description='Use 0 for no bandpass filtering')
    freq_max = mlpr.FloatParameter(
        optional=True, default=6000, description='Use 0 for no bandpass filtering')
    merge_thresh = mlpr.FloatParameter(
        optional=True, default=0.98, description='Threshold for automated merging')
    pc_per_chan = mlpr.IntegerParameter(
        optional=True, default=3, description='Number of principal components per channel')

    # added in version 0.2.4
    whiten = mlpr.BoolParameter(
        optional=True, default=False, description='Whether to do channel whitening as part of preprocessing')
    filter_type = mlpr.StringParameter(
        optional=True, default='bandpass', description='{none, bandpass, wiener, fftdiff, ndiff}')
    filter_detect_type = mlpr.StringParameter(
        optional=True, default='none', description='{none, bandpass, wiener, fftdiff, ndiff}')
    common_ref_type = mlpr.StringParameter(
        optional=True, default='none', description='{none, mean, median}')
    nTime_clu = mlpr.IntegerParameter(
        optional=True, default=1, description='Number of time periods to cluster together')
    nTime_drift = mlpr.IntegerParameter(
        optional=True, default=1, description='Number of time segments for drift correction')
    knn = mlpr.IntegerParameter(
        optional=True, default=30, description='K nearest neighbors')
    min_count = mlpr.IntegerParameter(
        optional=True, default=30, description='Minimum cluster size')
    fGpu = mlpr.BoolParameter(
        optional=True, default=True, description='Use GPU if available')
    fft_thresh = mlpr.FloatParameter(
        optional=True, default=0, description='FFT-based noise peak threshold')
    nSites_whiten = mlpr.IntegerParameter(
        optional=True, default=32, description='Number of adjacent channels to whiten')
    feature_type = mlpr.StringParameter(
        optional=True, default='gpca', description='{gpca, pca, vpp, vmin, vminmax, cov, energy, xcov}')

    def run(self):
        ironclust_path = os.environ.get('IRONCLUST_PATH', None)
        if not ironclust_path:
            raise Exception('Environment variable not set: IRONCLUST_PATH')

        code = ''.join(random.choice(string.ascii_uppercase)
                       for x in range(10))
        tmpdir = os.environ.get('TEMPDIR', '/tmp')+'/ironclust-tmp-'+code

        print('------------------------------------- using', tmpdir)

        try:
            recording = SFMdaRecordingExtractor(self.recording_dir)
            params = read_dataset_params(self.recording_dir)
            if len(self.channels) > 0:
                recording = se.SubRecordingExtractor(
                    parent_recording=recording, channel_ids=self.channels)
            if not os.path.exists(tmpdir):
                os.mkdir(tmpdir)

            all_params=dict()
            for param0 in self.PARAMETERS:
                all_params[param0.name] = getattr(self, param0.name)
            sorting = ironclust_helper(
                recording=recording,
                tmpdir=tmpdir,
                **all_params,
                # detect_sign=self.detect_sign,
                # adjacency_radius=self.adjacency_radius,
                # detect_threshold=self.detect_threshold,
                # merge_thresh=self.merge_thresh,
                # freq_min=self.freq_min,
                # freq_max=self.freq_max,
                # pc_per_chan=self.pc_per_chan,
                # whiten=self.whiten,
                # filter_type=self.filter_type,
                # filter_detect_type=self.filter_detect_type,
                # common_ref_type=self.common_ref_type,
                # nTime_clu=self.nTime_clu,
                # nTime_drift=self.nTime_drift,
                # knn=self.knn,
                # min_count=self.min_count,
                # fGpu=self.fGpu,
                # fft_thresh=self.fft_thresh,
                # nSites_whiten=self.nSites_whiten,
                # feature_type=self.feature_type,
                # prm_template_name=self.prm_template_name,
                ironclust_path=ironclust_path,
                params=params,
            )
            SFMdaSortingExtractor.writeSorting(
                sorting=sorting, save_path=self.firings_out)
        except:
            print('---------------------------------------', tmpdir, os.path.exists(tmpdir))
            raise
            if os.path.exists(tmpdir):
                if not getattr(self, '_keep_temp_files', False):
                    shutil.rmtree(tmpdir)
            raise
        if not getattr(self, '_keep_temp_files', False):
            pass
            # shutil.rmtree(tmpdir)


def ironclust_helper(*,
    recording,  # Recording object
    tmpdir,  # Temporary working directory
    # detect_sign=-1,  # Polarity of the spikes, -1, 0, or 1
    # adjacency_radius=-1,  # Channel neighborhood adjacency radius corresponding to geom file
    # detect_threshold=5,  # Threshold for detection
    # merge_thresh=.98,  # Cluster merging threhold 0..1
    # freq_min=300,  # Lower frequency limit for band-pass filter
    # freq_max=6000,  # Upper frequency limit for band-pass filter
    # pc_per_chan=3,  # Number of pc per channel
    # whiten=False,  # Spatial whitening
    # filter_type='bandpass',  # filter type
    # common_ref_type='none',  # common average reference type
    # nTime_clu=1,  # Number of time periods to cluster together
    # nTime_drift=1,  # Number of time segments for drift correction
    # knn=30,  # K nearest neighbors
    # min_count=30,  # minimum cluster size
    # fGpu=True,  # use GPU if available
    # fft_thresh=0,  # fft-cleanup threshold
    # nSites_whiten=32,  # number of adjacent channels to whiten
    # feature_type='gpca',  # feature name
    # prm_template_name='',  # Name of the template file
    ironclust_path=None,
    params=dict(),
    **kwargs
    ):
    if ironclust_path is None:
        ironclust_path = os.getenv('ironclust_path', None)
    if not ironclust_path:
        raise Exception(
            'You must either set the ironclust_path environment variable, or pass the ironclust_path parameter')
    # source_dir=os.path.dirname(os.path.realpath(__file__))

    dataset_dir = tmpdir+'/ironclust_dataset'
    # Generate three files in the dataset directory: raw.mda, geom.csv, params.json
    SFMdaRecordingExtractor.writeRecording(
        recording=recording, save_path=dataset_dir, params=params)

    samplerate = recording.getSamplingFrequency()

    print('Reading timeseries header...')
    HH = mdaio.readmda_header(dataset_dir+'/raw.mda')
    num_channels = HH.dims[0]
    num_timepoints = HH.dims[1]
    duration_minutes = num_timepoints/samplerate/60
    print('Num. channels = {}, Num. timepoints = {}, duration = {} minutes'.format(
        num_channels, num_timepoints, duration_minutes))

    print('Creating argfile.txt...')
    txt = ''
    for key0, val0 in kwargs.items():
        txt += '{}={}\n'.format(key0, val0)
    # txt += 'samplerate={}\n'.format(samplerate)
    # txt += 'detect_sign={}\n'.format(detect_sign)
    # txt += 'adjacency_radius={}\n'.format(adjacency_radius)
    # txt += 'detect_threshold={}\n'.format(detect_threshold)
    # txt += 'merge_thresh={}\n'.format(merge_thresh)
    # txt += 'freq_min={}\n'.format(freq_min)
    # txt += 'freq_max={}\n'.format(freq_max)
    # txt += 'pc_per_chan={}\n'.format(pc_per_chan)
    # txt += 'prm_template_name={}\n'.format(prm_template_name)
    # txt += 'whiten={}\n'.format(whiten)
    # txt += 'filter_type={}\n'.format(filter_type)
    # txt += 'filter_detect_type={}\n'.format(filter_detect_type)
    # txt += 'common_ref_type={}\n'.format(common_ref_type)
    # txt += 'nTime_clu={}\n'.format(nTime_clu)
    # txt += 'nTime_drift={}\n'.format(nTime_drift)
    # txt += 'knn={}\n'.format(knn)
    # txt += 'min_count={}\n'.format(min_count)
    # txt += 'fGpu={}\n'.format(fGpu)
    # txt += 'fft_thresh={}\n'.format(fft_thresh)
    # txt += 'nSites_whiten={}\n'.format(nSites_whiten)
    # txt += 'feature_type={}\n'.format(feature_type)

    if 'scale_factor' in params:
        txt += 'scale_factor={}\n'.format(params["scale_factor"])
    _write_text_file(dataset_dir+'/argfile.txt', txt)

    # new method
    print('Running ironclust in {tmpdir}...'.format(tmpdir=tmpdir))
    cmd = '''
        addpath('{ironclust_path}', '{ironclust_path}/matlab', '{ironclust_path}/matlab/mdaio');
        try
            p_ironclust('{tmpdir}', '{dataset_dir}/raw.mda', '{dataset_dir}/geom.csv', '', '', '{tmpdir}/firings.mda', '{dataset_dir}/argfile.txt');
        catch
            fprintf('----------------------------------------');
            fprintf(lasterr());
            quit(1);
        end
        quit(0);
    '''
    cmd = cmd.format(ironclust_path=ironclust_path, tmpdir=tmpdir, dataset_dir=dataset_dir)

    matlab_cmd = mlpr.ShellScript(cmd, script_path=tmpdir+'/run_ironclust.m',keep_temp_files=True)
    matlab_cmd.write()

    shell_cmd = '''
        #!/bin/bash
        cd {tmpdir}
        echo '=====================' `date` '=====================' >> run_ironclust.log 
        matlab -nosplash -nodisplay -r run_ironclust
    '''.format(tmpdir=tmpdir)
    shell_cmd = mlpr.ShellScript(shell_cmd, script_path=tmpdir+'/run_ironclust.sh', keep_temp_files=True)
    shell_cmd.write(tmpdir+'/run_ironclust.sh')
    shell_cmd.start()
    
    retcode = shell_cmd.wait()

    if retcode != 0:
        raise Exception('ironclust returned a non-zero exit code')

    # old method
    # print('Running IronClust...')
    # cmd_path = "addpath('{}', '{}/matlab', '{}/matlab/mdaio');".format(
    #     ironclust_path, ironclust_path, ironclust_path)
    # # "p_ironclust('$(tempdir)','$timeseries$','$geom$','$prm$','$firings_true$','$firings_out$','$(argfile)');"
    # cmd_call = "p_ironclust('{}', '{}', '{}', '', '', '{}', '{}');"\
    #     .format(tmpdir, dataset_dir+'/raw.mda', dataset_dir+'/geom.csv', tmpdir+'/firings.mda', dataset_dir+'/argfile.txt')
    # cmd = 'matlab -nosplash -nodisplay -r "{} {} quit;"'.format(
    #     cmd_path, cmd_call)
    # print(cmd)
    # #retcode = _run_command_and_print_output(cmd)
    # retcode = os.system(cmd)

    # if retcode != 0:
    #     raise Exception('IronClust returned a non-zero exit code')

    # parse output
    result_fname = tmpdir+'/firings.mda'
    if not os.path.exists(result_fname):
        raise Exception('Result file does not exist: ' + result_fname)

    firings = mdaio.readmda(result_fname)
    sorting = se.NumpySortingExtractor()
    sorting.setTimesLabels(firings[1, :], firings[2, :])
    return sorting


def _read_text_file(fname):
    with open(fname) as f:
        return f.read()


def _write_text_file(fname, str):
    with open(fname, 'w') as f:
        f.write(str)


def _run_command_and_print_output(command):
    with Popen(shlex.split(command), stdout=PIPE, stderr=PIPE) as process:
        while True:
            output_stdout = process.stdout.readline()
            output_stderr = process.stderr.readline()
            if (not output_stdout) and (not output_stderr) and (process.poll() is not None):
                break
            if output_stdout:
                print(output_stdout.decode())
            if output_stderr:
                print(output_stderr.decode())
        rc = process.poll()
        return rc


def read_dataset_params(dsdir):
    #ca = _load_required_modules()
    fname1 = dsdir+'/params.json'
    fname2 = mt.realizeFile(path=fname1)
    if not fname2:
        raise Exception('Unable to find file: '+fname1)
    if not os.path.exists(fname2):
        raise Exception('Dataset parameter file does not exist: ' + fname2)
    with open(fname2) as f:
        return json.load(f)



import time
from PIL import Image
import os
from copy import deepcopy
from mountaintools import client as mt
import mlprocessors as mlpr
import multiprocessing
# from matplotlib import pyplot as plt
from .compute_units_info import ComputeUnitsInfo
from .computerecordinginfo import ComputeRecordingInfo
import mtlogging

@mtlogging.log()
def summarize_recordings(recordings, compute_resource=None, label=None):
    print('')
    print('>>>>>> {}'.format(label or 'summarize recordings'))

    jobs_info = ComputeRecordingInfo.createJobs([
        dict(
          recording_dir=recording['directory'],
          channels=recording.get('channels',[]),
          json_out={'ext':'.json','upload':True},
          _container='default',
          _label='Summarize recording: '+recording.get('name', '')
        )
        for recording in recordings
    ])

    jobs_units_info = ComputeUnitsInfo.createJobs([
        dict(
          recording_dir=recording['directory'],
          firings=recording['directory']+'/firings_true.mda',
          unit_ids=recording.get('units_true',None),
          channel_ids=recording.get('channels',None),
          json_out={'ext':'.json','upload':True},
          _container='default',
          _label='Compute units info for recording: '+recording.get('name', '')
        )
        for recording in recordings
    ])
    
    # all_jobs=jobs_info+jobs_timeseries_plot+jobs_units_info
    all_jobs=jobs_info+jobs_units_info
    label=label or 'Summarize recordings'
    mlpr.executeBatch(jobs=all_jobs,label=label,num_workers=None,compute_resource=compute_resource)

    print('Gathering summarized recordings...')

    summarized_recordings = deepcopy(recordings)
    for ii in range(len(recordings)):
      summary = dict()

      result0 = jobs_info[ii].result
      summary['computed_info'] = mt.loadObject(path=result0.outputs['json_out'])
      summary['plots']=dict()
    
      result0=jobs_units_info[ii].result
      summary['true_units_info']=mt.getSha1Url(path=result0.outputs['json_out'], basename='true_units_info.json')
      
      summarized_recordings[ii]['summary'] = summary

    return summarized_recordings

def _create_jobs_for_recording_old(recording):
    print('Creating jobs for recording: {}/{}'.format(recording.get('study',''),recording.get('name','')))
    dsdir=recording['directory']
    # raw_path=dsdir+'/raw.mda'
    firings_true_path=dsdir+'/firings_true.mda'
    # geom_path=dsdir+'/geom.csv'
    channels=recording.get('channels',None)
    units=recording.get('units_true',None)

    if not mt.findFile(path=firings_true_path):
        raise Exception('firings_true file not found: '+firings_true_path)
    job_info=ComputeRecordingInfo.createJob(
        recording_dir=dsdir,
        channels=recording.get('channels',[]),
        json_out={'ext':'.json','upload':True},
        _container='default'
    )
    job_info.addFilesToRealize([dsdir+'/raw.mda',dsdir+'/firings_true.mda'])
    # job=CreateTimeseriesPlot.createJob(
    #     recording_dir=dsdir,
    #     channels=recording.get('channels',[]),
    #     jpg_out={'ext':'.jpg','upload':True},
    #     _container='default'
    # )
    # jobs_timeseries_plot.append(job)
    job_units_info=ComputeUnitsInfo.createJob(
        recording_dir=dsdir,
        firings=dsdir+'/firings_true.mda',
        unit_ids=units,
        channel_ids=channels,
        json_out={'ext':'.json','upload':True},
        _container='default'
    )
    job_units_info.addFilesToRealize([dsdir+'/raw.mda', dsdir+'/firings_true.mda'])
    return (job_info, job_units_info)


import mlprocessors as mlpr
import json

import spikeextractors as si
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor

_CONTAINER='sha1://1ad2478736ad188ab5050289ffb1d2c29d1ba750/03-29-2019/spikeforest_basic.simg'


# A MountainLab processor for generating the summary info for a recording
class ComputeRecordingInfo(mlpr.Processor):
  NAME='ComputeRecordingInfo'
  VERSION='0.1.1'
  CONTAINER=_CONTAINER

  recording_dir=mlpr.Input(directory=True,description='Recording directory')
  channels=mlpr.IntegerListParameter(description='List of channels to use.',optional=True,default=[])
  json_out=mlpr.Output('Info in .json file')
    
  def run(self):
    ret={}
    recording=SFMdaRecordingExtractor(dataset_directory=self.recording_dir,download=True)
    if len(self.channels)>0:
      recording=si.SubRecordingExtractor(parent_recording=recording,channel_ids=self.channels)
    ret['samplerate']=recording.getSamplingFrequency()
    ret['num_channels']=len(recording.getChannelIds())
    ret['duration_sec']=recording.getNumFrames()/ret['samplerate']
    write_json_file(self.json_out,ret)

def write_json_file(fname,obj):
  with open(fname, 'w') as f:
    json.dump(obj, f)



from .mountainsort4 import MountainSort4, MountainSort4TestError
from .spyking_circus import SpykingCircus
from .yass import YASS
from .ironclust import IronClust
from .kilosort import KiloSort
from .kilosort2 import KiloSort2
#from .klusta import klusta



import vdomr as vd
import time
import multiprocessing
import sys
from .stdoutsender import StdoutSender
import mtlogging
import numpy as np
from copy import deepcopy
import spikeforest_analysis as sa
from spikeforest import EfficientAccessRecordingExtractor
import spikeforestwidgets as SFW
import json
from matplotlib import pyplot as plt
import mlprocessors as mlpr
import uuid

class TemplatesView(vd.Component):
    def __init__(self, *, context, opts=None, prepare_result):
        vd.Component.__init__(self)
        self._sorting_context = context
        self._recording_context = context.recordingContext()
        self._size = (100, 100)
        units = prepare_result['units']
        self._templates_widget = TemplatesWidget(units=units)
        self._update_selected()
        self._templates_widget.onSelectedUnitIdsChanged(lambda: self._sorting_context.setSelectedUnitIds(self._templates_widget.selectedUnitIds()))
        self._sorting_context.onSelectedUnitIdsChanged(lambda: self._templates_widget.setSelectedUnitIds(self._sorting_context.selectedUnitIds()))
        self._templates_widget.onCurrentUnitIdChanged(lambda: self._sorting_context.setCurrentUnitId(self._templates_widget.currentUnitId()))
        self._sorting_context.onCurrentUnitIdChanged(lambda: self._templates_widget.setCurrentUnitId(self._sorting_context.currentUnitId()))
        #self._templates_widget.setSize(self._size)
        self.refresh()
    @staticmethod
    def prepareView(context, opts):
        sorting_context = context
        recording_context = context.recordingContext()
        sorting_context.initialize()
        earx = EfficientAccessRecordingExtractor(recording=recording_context.recordingExtractor())
        sorting = sorting_context.sortingExtractor()
        unit_ids = sorting.getUnitIds()
        print('***** Computing unit templates...')
        #templates = compute_unit_templates(recording=earx, sorting=sorting, unit_ids=unit_ids)
        templates = ComputeUnitTemplates.execute(recording=earx, sorting=sorting, unit_ids=unit_ids, templates_out=True).outputs['templates_out']
        print('*****')
        return dict(units=[
            dict(
                template=templates[:,:,i],
                unit_id=unit_id
            )
            for i, unit_id in enumerate(unit_ids)
        ])
    def setSize(self, size):
        if self._size != size:
            self._size=size
        #if self._templates_widget:
        #    self._templates_widget.setSize(size)

    def size(self):
        return self._size
    def tabLabel(self):
        return 'Templates'
    def render(self):
        return self._templates_widget

    def _update_selected(self):
        self._templates_widget.setSelectedUnitIds(self._sorting_context.selectedUnitIds())
        self._templates_widget.setCurrentUnitId(self._sorting_context.currentUnitId())

def get_random_spike_waveforms(*,recording,sorting,unit,snippet_len,max_num,channels=None):
    st=sorting.getUnitSpikeTrain(unit_id=unit)
    num_events=len(st)
    if num_events>max_num:
        event_indices=np.random.choice(range(num_events),size=max_num,replace=False)
    else:
        event_indices=range(num_events)

    spikes=recording.getSnippets(reference_frames=st[event_indices].astype(int),snippet_len=snippet_len,channel_ids=channels)
    if len(spikes)>0:
      spikes=np.dstack(tuple(spikes))
    else:
      spikes=np.zeros((recording.getNumChannels(),snippet_len,0))
    return spikes

def compute_unit_templates(*,recording,sorting,unit_ids,snippet_len=50,max_num=100,channels=None):
    M = len(recording.getChannelIds())
    T = snippet_len
    K = len(unit_ids)
    ret = np.zeros((M,T,K))
    for i, unit in enumerate(unit_ids):
        print('Unit {} of {} (id={})'.format(i, len(unit_ids), unit))
        waveforms = get_random_spike_waveforms(recording=recording, sorting=sorting, unit=unit, snippet_len=snippet_len, max_num=max_num, channels=None)
        template = np.median(waveforms,axis=2)
        ret[:,:,i] = template
    
    return ret


class ComputeUnitTemplates(mlpr.Processor):
    NAME = 'ComputeUnitTemplates'
    VERSION = '0.1.4'
    recording = mlpr.Input()
    sorting = mlpr.Input()
    templates_out = mlpr.OutputArray()

    def run(self):
        templates = compute_unit_templates(recording=self.recording, sorting=self.sorting, unit_ids=self.sorting.getUnitIds()) # pylint: disable=no-member
        print('Saving templates...')
        np.save(self.templates_out, templates)
        

class TemplatesWidget(vd.Component):
  def __init__(self,*,units):
    vd.Component.__init__(self)
    y_scale_factor = _compute_initial_y_scale_factor(units)
    self._current_unit_id = None
    self._selected_unit_ids = []
    self._selected_unit_ids_changed_handlers = []
    self._current_unit_id_changed_handlers = []
    self._widgets=[
        TemplateWidget(
            template=unit['template'],
            unit_id=unit['unit_id'],
            y_scale_factor=y_scale_factor,
            size=(150,300)
        )
        for unit in units
    ]
    self._uuid=str(uuid.uuid4())
    vd.register_callback(self._uuid+'_box_clicked', self._handle_box_clicked)
    vd.devel.loadBootstrap()
  def setSelectedUnitIds(self,ids):
    if ids is None:
        ids=[]
    ids = sorted([int(id) for id in ids])
    if ids == self._selected_unit_ids:
        return
    self._selected_unit_ids = ids
    ids_set=set(ids)
    for W in self._widgets:
        W.setSelected(W.unitId() in ids_set)
    for handler in self._selected_unit_ids_changed_handlers:
        handler()

  def selectedUnitIds(self):
      return deepcopy(self._selected_unit_ids)

  def onSelectedUnitIdsChanged(self, handler):
      self._selected_unit_ids_changed_handlers.append(handler)

  def setCurrentUnitId(self,id):
    if id == self._current_unit_id:
        return
    self._current_unit_id = id
    for W in self._widgets:
        W.setCurrent(W.unitId() == id)
    for handler in self._current_unit_id_changed_handlers:
        handler()

  def currentUnitId(self):
      return self._current_unit_id

  def onCurrentUnitIdChanged(self, handler):
      self._current_unit_id_changed_handlers.append(handler)
    
  def _handle_box_clicked(self, *, index, ctrlKey):
      W = self._widgets[int(index)]
      if ctrlKey:
          a = self.selectedUnitIds()
          if W.unitId() in a:
              a.remove(W.unitId())
          else:
              a.append(W.unitId())
          self.setSelectedUnitIds(a)
      else:
        self.setSelectedUnitIds([W.unitId()])
        self.setCurrentUnitId(W.unitId())

  def render(self):
    box_style=dict(float='left')
    boxes=[
        vd.div(W,id=self._uuid+'-{}'.format(i),style=box_style)
        for i,W in enumerate(self._widgets)
    ]
    div=vd.div(
        boxes,
        style=dict(overflow='auto',height='100%')
    )
    return div

  def postRenderScript(self):
      js="""
      for (let i=0; i<{num_widgets}; i++) {
          (function(index) {
            let elmt = document.getElementById('{uuid}-'+index);
            if (elmt) {
                elmt.onclick=function(evt) {
                    window.vdomr_invokeFunction('{box_clicked_callback_id}', [], {index:index, ctrlKey:evt.ctrlKey});
                }
            }
          })(i);
      }
      """
      js=js.replace('{num_widgets}', str(len(self._widgets)))
      js=js.replace('{uuid}', self._uuid)
      js=js.replace('{box_clicked_callback_id}', self._uuid+'_box_clicked')
      return js

def _compute_initial_y_scale_factor(units):
    templates_list = [unit['template'] for unit in units]
    all_templates = np.stack(templates_list)
    all_abs_vals = np.abs(all_templates.ravel())
    vv = np.percentile(all_abs_vals, 99)
    if vv>0:
        return 1/(2*vv)
    else:
        return 1
class TemplateWidget(vd.Component):
  def __init__(self,*,unit_id,template,y_scale_factor=None,size=(200,200)):
    vd.Component.__init__(self)
    self._unit_id=unit_id
    self._selected=False
    self._current=False
    self._size=size
    self._size_plot=(size[0]-20, size[1]-40)
    self._plot=SFW.TemplateWidget(template=template,size=self._size_plot)
    self._plot.setYScaleFactor(y_scale_factor)
  def setSelected(self,val):
    if self._selected==val:
        return
    self._selected=val
    self.refresh()
  def setCurrent(self,val):
    if self._current==val:
        return
    self._current=val
    self.refresh()
  def unitId(self):
    return self._unit_id
  def render(self):
    style_plot={'position':'relative', 'border':'solid 1px black', 'background-color':'white', 'left':'10px', 'bottom':'10px', 'width':'{}px'.format(self._size_plot[0]), 'height':'{}px'.format(self._size_plot[1])}
    style1={'width':'{}px'.format(self._size[0]), 'height':'{}px'.format(self._size[1])}
    if self._selected:
        style1['background-color']='rgb(240,240,200)'
        style1['border']='solid 1px gray'
        if self._current:
            style1['background-color']='rgb(220,220,180)'
    if self._current:
        style1['border']='solid 2px rgb(150,50,50)'
    return vd.div(
        vd.p('Unit {}'.format(self._unit_id),style={'text-align':'center'}),
        vd.div(self._plot,style=style_plot,size=()),
        style=style1
    )





import mlprocessors as mlpr
import os
import time
import numpy as np
from os.path import join
from subprocess import Popen, PIPE
import shlex
import random
import string
import shutil
from spikeforest import mdaio
import spikeextractors as se
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
import sys
import shlex
#import h5py


class KiloSort2(mlpr.Processor):
    NAME = 'KiloSort2'
    VERSION = '0.2.1'  # wrapper VERSION
    ADDITIONAL_FILES = ['*.m']
    ENVIRONMENT_VARIABLES = [
        'NUM_WORKERS', 'MKL_NUM_THREADS', 'NUMEXPR_NUM_THREADS', 'OMP_NUM_THREADS']
    CONTAINER = None
    CONTAINER_SHARE_ID = None

    recording_dir = mlpr.Input('Directory of recording', directory=True)
    channels = mlpr.IntegerListParameter(
        description='List of channels to use.', optional=True, default=[])
    firings_out = mlpr.Output('Output firings file')

    detect_sign = mlpr.IntegerParameter(
        'Use -1 or 1, depending on the sign of the spikes in the recording')
    adjacency_radius = mlpr.FloatParameter(
        'Use -1 to include all channels in every neighborhood')
    detect_threshold = mlpr.FloatParameter(
        optional=True, default=3, description='')
    # prm_template_name=mlpr.StringParameter(optional=False,description='TODO')
    freq_min = mlpr.FloatParameter(
        optional=True, default=150, description='Use 0 for no bandpass filtering')
    freq_max = mlpr.FloatParameter(
        optional=True, default=6000, description='Use 0 for no bandpass filtering')
    merge_thresh = mlpr.FloatParameter(
        optional=True, default=0.98, description='TODO')
    pc_per_chan = mlpr.IntegerParameter(
        optional=True, default=3, description='TODO')

    def run(self):
        code = ''.join(random.choice(string.ascii_uppercase)
                       for x in range(10))
        tmpdir = os.environ.get('TEMPDIR', '/tmp')+'/kilosort2-tmp-'+code

        try:
            recording = SFMdaRecordingExtractor(self.recording_dir)
            if len(self.channels) > 0:
                recording = se.SubRecordingExtractor(
                    parent_recording=recording, channel_ids=self.channels)
            if not os.path.exists(tmpdir):
                os.mkdir(tmpdir)
            sorting = kilosort2_helper(
                recording=recording,
                tmpdir=tmpdir,
                detect_sign=self.detect_sign,
                adjacency_radius=self.adjacency_radius,
                detect_threshold=self.detect_threshold,
                merge_thresh=self.merge_thresh,
                freq_min=self.freq_min,
                freq_max=self.freq_max,
                pc_per_chan=self.pc_per_chan
            )
            SFMdaSortingExtractor.writeSorting(
                sorting=sorting, save_path=self.firings_out)
        except:
            #if os.path.exists(tmpdir):
            #    shutil.rmtree(tmpdir)
            raise
        #shutil.rmtree(tmpdir)


def kilosort2_helper(*,
                    recording,  # Recording object
                    tmpdir,  # Temporary working directory
                    detect_sign=-1,  # Polarity of the spikes, -1, 0, or 1
                    adjacency_radius=-1,  # Channel neighborhood adjacency radius corresponding to geom file
                    detect_threshold=5,  # Threshold for detection
                    merge_thresh=.98,  # Cluster merging threhold 0..1
                    freq_min=150,  # Lower frequency limit for band-pass filter
                    freq_max=6000,  # Upper frequency limit for band-pass filter
                    pc_per_chan=3,  # Number of pc per channel
                    KILOSORT2_PATH=None,  # github kilosort2
                    IRONCLUST_PATH=None  # github ironclust
                    ):
    if KILOSORT2_PATH is None:
        KILOSORT2_PATH = os.getenv('KILOSORT2_PATH', None)
    if not KILOSORT2_PATH:
        raise Exception(
            'You must either set the KILOSORT2_PATH environment variable, or pass the KILOSORT2_PATH parameter')

    if IRONCLUST_PATH is None:
        IRONCLUST_PATH = os.getenv('IRONCLUST_PATH', None)
    if not IRONCLUST_PATH:
        raise Exception(
            'You must either set the IRONCLUST_PATH environment variable, or pass the IRONCLUST_PATH parameter')

    source_dir = os.path.dirname(os.path.realpath(__file__))

    dataset_dir = tmpdir+'/kilosort2_dataset'
    # Generate three files in the dataset directory: raw.mda, geom.csv, params.json
    SFMdaRecordingExtractor.writeRecording(
        recording=recording, save_path=dataset_dir)

    samplerate = recording.getSamplingFrequency()

    print('Reading timeseries header...')
    HH = mdaio.readmda_header(dataset_dir+'/raw.mda')
    num_channels = HH.dims[0]
    num_timepoints = HH.dims[1]
    duration_minutes = num_timepoints/samplerate/60
    print('Num. channels = {}, Num. timepoints = {}, duration = {} minutes'.format(
        num_channels, num_timepoints, duration_minutes))

    print('Creating argfile.txt file...')
    txt = ''
    txt += 'samplerate={}\n'.format(samplerate)
    txt += 'detect_sign={}\n'.format(detect_sign)
    txt += 'adjacency_radius={}\n'.format(adjacency_radius)
    txt += 'detect_threshold={}\n'.format(detect_threshold)
    txt += 'merge_thresh={}\n'.format(merge_thresh)
    txt += 'freq_min={}\n'.format(freq_min)
    txt += 'freq_max={}\n'.format(freq_max)
    txt += 'pc_per_chan={}\n'.format(pc_per_chan)
    _write_text_file(dataset_dir+'/argfile.txt', txt)

    print('Running kilosort2...')
    cmd_path = "addpath('{}'); ".format(source_dir)
    cmd_call = "p_kilosort2('{}', '{}', '{}', '{}', '{}', '{}', '{}');"\
        .format(KILOSORT2_PATH, IRONCLUST_PATH, tmpdir, dataset_dir+'/raw.mda', dataset_dir+'/geom.csv', tmpdir+'/firings.mda', dataset_dir+'/argfile.txt')
    cmd = 'matlab -nosplash -nodisplay -r "{} {} quit;"'.format(
        cmd_path, cmd_call)
    print(cmd)
    retcode = _run_command_and_print_output(cmd)

    if retcode != 0:
        raise Exception('kilosort2 returned a non-zero exit code')

    # parse output
    result_fname = tmpdir+'/firings.mda'
    if not os.path.exists(result_fname):
        raise Exception('Result file does not exist: ' + result_fname)

    firings = mdaio.readmda(result_fname)
    sorting = se.NumpySortingExtractor()
    sorting.setTimesLabels(firings[1, :], firings[2, :])
    return sorting


def _read_text_file(fname):
    with open(fname) as f:
        return f.read()


def _write_text_file(fname, str):
    with open(fname, 'w') as f:
        f.write(str)


def _run_command_and_print_output(command):
    with Popen(shlex.split(command), stdout=PIPE, stderr=PIPE) as process:
        while True:
            output_stdout = process.stdout.readline()
            output_stderr = process.stderr.readline()
            if (not output_stdout) and (not output_stderr) and (process.poll() is not None):
                break
            if output_stdout:
                print(output_stdout.decode())
            if output_stderr:
                print(output_stderr.decode())
        rc = process.poll()
        return rc



import setuptools

pkg_name = "spikeforest"

# NOTE: you should install this project in development mode
# > python setup.py develop

setuptools.setup(
    name=pkg_name,
    version="0.2.7",
    author="Jeremy Magland",
    author_email="jmagland@flatironinstitute.org",
    description="Spike sorting",
    packages=setuptools.find_packages(),
    package_dir={
        'spikesorters': 'spikesorters',
        'spikeforestwidgets': 'spikeforestwidgets',
        'forestview': 'forestview'},
    package_data={
        'spikesorters': ['spyking_circus/*.params', 'yass/config*'],
        'spikeforestwidgets': ['*.js','timeserieswidget/*.js', 'dist/*.js'],
        'forestview': ['*.js','*/*.js']
        },
    install_requires=[
        'numpy','scipy','matplotlib','requests','pillow','pandas','ipython','h5py','setuptools-git','scikit-learn',
        'spikeextractors>=0.3,<0.4'
    ],
    scripts=['bin/forestview'],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    )
)



import numpy as np
from matplotlib import pyplot as plt
import vdomr as vd

class UnitWaveformsWidget(vd.Component):
  def __init__(self,*,recording,sorting,max_num_spikes_per_unit=20,snippet_len=100):
    vd.Component.__init__(self)
    self._widgets=[
        UnitWaveformWidget(
            recording=recording,
            sorting=sorting,
            unit_id=id,
            average_waveform=None,
            max_num_spikes_per_unit=max_num_spikes_per_unit,
            snippet_len=snippet_len
        )
        for id in sorting.getUnitIds()
    ]
    vd.devel.loadBootstrap()
  def setSelectedUnitIds(self,ids):
    ids=set(ids)
    for W in self._widgets:
        W.setSelected(W.unitId() in ids)
  def render(self):
    box_style=dict(float='left')
    boxes=[
        vd.div(W,style=box_style)
        for W in self._widgets
    ]
    div=vd.div(boxes)
    return div

class UnitWaveformWidget(vd.Component):
  def __init__(self,*,recording,sorting,unit_id,average_waveform=None,show_average=True,max_num_spikes_per_unit=20,snippet_len=100):
    vd.Component.__init__(self)
    self._plot=_UnitWaveformPlot(
        recording=recording,
        sorting=sorting,
        unit_id=unit_id,
        average_waveform=average_waveform,
        show_average=show_average,
        max_num_spikes_per_unit=max_num_spikes_per_unit,
        snippet_len=snippet_len
    )
    self._plot_div=vd.components.LazyDiv(self._plot)
    self._unit_id=unit_id
    self._selected=False
  def setSelected(self,val):
    if self._selected==val:
        return
    self._selected=val
    self.refresh()
  def unitId(self):
    return self._unit_id
  def render(self):
    style0={'border':'solid 1px black','margin':'5px'}
    style1={}
    if self._selected:
        style1['background-color']='yellow'
    return vd.div(
        vd.p('Unit {}'.format(self._unit_id),style={'text-align':'center'}),
        vd.div(self._plot_div,style=style0),
        style=style1
    )

import time
from spikeforest import spikewidgets as sw
class _UnitWaveformPlot(vd.components.Pyplot):
  def __init__(self,*,recording,sorting,unit_id,average_waveform,show_average,max_num_spikes_per_unit,snippet_len):
    vd.components.Pyplot.__init__(self)
    self._recording=recording
    self._sorting=sorting
    self._unit_id=unit_id
    self._max_num_spikes_per_unit=max_num_spikes_per_unit
    self._average_waveform=average_waveform
    self._show_average=show_average
    self._snippet_len=snippet_len
  def plot(self):
    #W=sw.UnitWaveformsWidget(recording=self._recording,sorting=self._sorting,unit_ids=[self._unit_id],width=5,height=5)
    #W.plot()
    plot_unit_waveform(
        recording=self._recording,
        sorting=self._sorting,
        unit_id=self._unit_id,
        average_waveform=self._average_waveform,
        show_average=self._show_average,
        max_num_spikes_per_unit=self._max_num_spikes_per_unit,
        snippet_len=self._snippet_len
    )

def _compute_minimum_gap(x):
  a=np.sort(np.unique(x))
  if len(a)<=1:
    return 1
  return np.min(np.diff(a))

def _plot_spike_shapes(*, representative_waveforms=None, average_waveform=None, show_average, channel_locations=None,
                           ylim=None, max_representatives=None, color='blue', title=''):
    if average_waveform is None:
        if representative_waveforms is None:
            raise Exception('You must provide either average_waveform, representative waveforms, or both')
        average_waveform = np.mean(representative_waveforms, axis=2)
    M = average_waveform.shape[0]  # number of channels
    T = average_waveform.shape[1]  # number of timepoints
    
    if ylim is None:
        ylim = [average_waveform.min(), average_waveform.max()]
    yrange = ylim[1] - ylim[0]
    
    if channel_locations is None:
        channel_locations = np.zeros((M, 2))
        for m in range(M):
            channel_locations[m, :] = [0, -m]

    if channel_locations.shape[1]>2:
      channel_locations=channel_locations[:,-2:]
    
    xmin=np.min(channel_locations[:,0])
    xmax=np.max(channel_locations[:,0])
    ymin=np.min(channel_locations[:,1])
    ymax=np.max(channel_locations[:,1])
    xgap=_compute_minimum_gap(channel_locations[:,0])
    ygap=_compute_minimum_gap(channel_locations[:,1])

    xvals = np.linspace(-xgap*0.8 / 2, xgap*0.8 / 2, T)
    yscale=1/(yrange/2)*ygap/2*0.4
    
    ax = plt.axes([0,0,1,1], frameon=False)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.get_xaxis().set_ticks([])
    ax.get_yaxis().set_ticks([])
    
    if representative_waveforms is not None:
        if max_representatives is not None:
            W0 = representative_waveforms
            if W0.shape[2] > max_representatives:
                indices = np.random.choice(range(W0.shape[2]), size=max_representatives, replace=False)
                representative_waveforms = W0[:, :, indices]
        L = representative_waveforms.shape[2]
        # for j in range(L):
        #     XX = np.zeros((T, M))
        #     YY = np.zeros((T, M))
        #     for m in range(M):
        #         loc = channel_locations[m, -2:]
        #         XX[:, m] = loc[0] + xvals
        #         YY[:, m] = loc[1] + (representative_waveforms[m, :, j] - representative_waveforms[m, 0, j])*yscale
        #     color=(np.random.uniform(0,1), np.random.uniform(0,1), np.random.uniform(0,1))
        #     plt.plot(XX, YY, color=color, alpha=0.3)    
        XX = np.zeros((T, M, L))
        YY = np.zeros((T, M, L))
        for m in range(M):
            loc = channel_locations[m, -2:]
            for j in range(L):
                XX[:, m, j] = loc[0] + xvals
                YY[:, m, j] = loc[1] + (representative_waveforms[m, :, j] - representative_waveforms[m, 0, j])*yscale
        XX = XX.reshape(T, M * L)
        YY = YY.reshape(T, M * L)
        plt.plot(XX, YY, color=(0.5, 0.5, 0.5), alpha=0.5)

        if show_average:
            XX = np.zeros((T, M))
            YY = np.zeros((T, M))
            for m in range(M):
                loc = channel_locations[m, -2:]
                XX[:, m] = loc[0] + xvals
                YY[:, m] = loc[1] + (average_waveform[m, :] - average_waveform[m, 0])*yscale
            plt.plot(XX, YY, color)
        
    plt.xlim(xmin-xgap/2,xmax+xgap/2)
    plt.ylim(ymin-ygap/2,ymax+ygap/2)
    
    #plt.gca().set_axis_off()
    if title:
        plt.title(title, color='gray')
        
def _get_random_spike_waveforms(*,recording, sorting, unit, max_num, channels, snippet_len):
    st = sorting.getUnitSpikeTrain(unit_id=unit)
    num_events = len(st)
    if num_events > max_num:
        event_indices = np.random.choice(range(num_events), size=max_num, replace=False)
    else:
        event_indices = range(num_events)

    spikes = recording.getSnippets(reference_frames=st[event_indices].astype(int), snippet_len=snippet_len,
                                  channel_ids=channels)
    if len(spikes)>0:
        spikes = np.dstack(tuple(spikes))
    else:
        spikes = np.zeros((recording.getNumChannels(), snippet_len, 0))
    return spikes

def plot_unit_waveform(*,recording,sorting,unit_id,max_num_spikes_per_unit,average_waveform,show_average,channel_ids=None,snippet_len=100,title=''):
  if not channel_ids:
    channel_ids=recording.getChannelIds()
  M = len(channel_ids)
  channel_locations = np.zeros((M, 2))
  for ii, ch in enumerate(channel_ids):
      loc = recording.getChannelProperty(ch, 'location')
      channel_locations[ii, :] = loc[-2:]
  
  spikes = _get_random_spike_waveforms(recording=recording,sorting=sorting,unit=unit_id, max_num=max_num_spikes_per_unit, channels=channel_ids, snippet_len=snippet_len)
  #if not title:
  #  title='Unit {}'.format(int(unit_id))
  
  _plot_spike_shapes(representative_waveforms=spikes,channel_locations=channel_locations, average_waveform=average_waveform, show_average=show_average, title=title)
